package de.heinerkuecker.asciiartflowparser;

import java.util.ArrayList;

import de.heinerkuecker.util.StringUtil;

/**
 * Diese Klasse reprsentiert eine Kreuzung
 * zweier Verbindungen (FlowConnection).
 *
 * Es gibt horizontale Kreuzungen
 * <pre>
 *   |
 *  -|-
 *   |
 * </pre>
 *
 * und vertikale Kreuzungen
 * <pre>
 *   |
 *  ---
 *   |
 * </pre>
 *
 * Die Kreuzung bezieht sich immer auf die
 * unterbrochene Linie, whrend die nicht
 * unterbrochene gekreuzte Linie kein
 * FlowConnectionCrossing-Objekt enthlt.
 *
 * @author Heiner Kcker
 */
public final class FlowConnectionCrossing
extends FlowConnectionPart
{
    private final int row;
    private final int col;
    private final int direction;
    private FlowConnectionLinePart crossedPart;

    /**
     * @return the col
     */
    int getCol()
    {
        return this.col;
    }

    /**
     * @return the direction
     */
    int getDirection()
    {
        return this.direction;
    }

    /**
     * @return the row
     */
    int getRow()
    {
        return this.row;
    }

    /**
     * @return the crossedPart
     */
    FlowConnectionLinePart getCrossedPart()
    {
        return this.crossedPart;
    }

    /**
     * @param pCrossedPart the crossedPart to set
     */
    void setCrossedPart(FlowConnectionLinePart pCrossedPart)
    {
        this.crossedPart = pCrossedPart;
    }

    /**
     * Constructor.
     *
     * @param pRow
     * @param pCol
     * @param pArrowChar
     */
    private FlowConnectionCrossing(
            final int pRow,
            final int pCol,
            final int pDirection,
            final FlowConnection pFlowConnection,
            final ArrayList pFlowConnectionCrossingList)
    {
        super( pFlowConnection );
        this.row = pRow;
        this.col = pCol;
        this.direction = pDirection;

        // vermerken der Kreuzung fr spteres Prfen, ob die Kreuzung evtl. ein Text war
        pFlowConnectionCrossingList.add( this );
    }

    /**
     * Zurckgeben der Zeile der
     * nchsten Position, wenn die Linie
     * weiter laufen wrde.
     *
     * @return
     */
    /*public*/ int getNextPosRow()
    {
        if ( this.direction == FlowConnection.DOWN )
            // nach unten
        {
            // eine Zeile nach unten
            return this.row + 1;
        }
        else if ( this.direction == FlowConnection.UP )
            // nach oben
        {
            // eine Zeile nach oben
            return this.row - 1;
        }
        return this.row;
    }

    /**
     * Zurckgeben der Spalte der
     * nchsten Position, wenn die Linie
     * weiter laufen wrde.
     *
     * @return
     */
    /*public*/ int getNextPosCol()
    {
        if ( this.direction == FlowConnection.RIGHT )
            // nach rechts
        {
            // eine Spalte nach rechts
            return this.col + 1;
        }
        else if ( this.direction == FlowConnection.LEFT )
        {
            // eine Spalte nach links
            return this.col - 1;
        }
        return this.col;
    }

    /**
     * @see FlowConnectionPart#matchPosition(int, int)
     */
    /*public*/ boolean matchPosition(
            final int pRowNr,
            final int pColNr)
    {
        return //--
            pRowNr == this.row
            && pColNr == this.col;
    }

    /**
     * debug output
     */
    public String toString()
    {
        return "\n\t\t" + StringUtil.getShortClassName( this ) + " " //--
            + "row: " + this.row + " " //--
            + "col: " + this.col + " " //--
            + "direction: " + FlowConnection.getDirectionStr( this.direction );
    }

    /**
     * static methods
     */

    /**
     * parse
     *
     * @param pFlowStrArr
     * @param pRow
     * @param pCol
     * @return
     */
    /*public*/ static FlowConnectionCrossing parse(
            final String[] pFlowStrArr,
            final int pRow,
            final int pCol,
            final int pDirection,
            final FlowConnection pFlowConnection,
            final ArrayList pFlowConnectionCrossingList )
    {
        switch ( pDirection )
        {
            case FlowConnection.UP:
                if ( pRow < 1 //--
                        || pCol < 2 //--
                        || pCol >= pFlowStrArr[ pRow ].length() //--
                        || pFlowStrArr[ pRow ].charAt( pCol ) != '-' //--
                        || pCol >= pFlowStrArr[ pRow - 1 ].length() //--
                        || pFlowStrArr[ pRow - 1 ].charAt( pCol ) != '|' )
                {
                    // keine Kreuzung
                    return null;
                }
                break;
            case FlowConnection.DOWN:
                if ( pRow + 1 > pFlowStrArr.length //--
                        || pCol < 2 //--
                        || pCol - 1 >= pFlowStrArr[ pRow ].length() //--
                        || pFlowStrArr[ pRow ].charAt( pCol - 1 ) != '-' //--
                        || pCol + 1 >= pFlowStrArr[ pRow ].length() //--
                        || pFlowStrArr[ pRow ].charAt( pCol + 1 ) != '-' )
                {
                    // keine Kreuzung
                    return null;
                }
                break;
            case FlowConnection.LEFT:
            {
                final String lineStr = pFlowStrArr[ pRow ];
                if ( pCol < 1 //--
                        || pCol >= pFlowStrArr[ pRow ].length() //--
                        || pFlowStrArr[ pRow ].charAt( pCol ) != '|' //--
                        || pCol >= lineStr.length() //--
                        || lineStr.charAt( pCol - 1 ) != '-' )
                {
                    // keine Kreuzung
                    return null;
                }
                break;
            }
            case FlowConnection.RIGHT:
            {
                final String lineStr = pFlowStrArr[ pRow ];
                if ( pRow < 1
                        //|| pCol < 0 //--
                        || pCol >= lineStr.length() //--
                        || lineStr.charAt( pCol ) != '|' //--
                        //|| pCol >= pFlowStrArr[ pRow - 1 ].length() //--
                        //|| pFlowStrArr[ pRow - 1 ].charAt( pCol ) != '|' //--
                        || pCol >= lineStr.length() - 1 //--
                        || lineStr.charAt( pCol + 1 ) != '-' )
                {
                    // keine Kreuzung
                    return null;
                }
                break;
            }
            default:
                throw new IllegalArgumentException( "unknown direction: " + pDirection );
        }
        return new FlowConnectionCrossing(
                pRow,
                pCol,
                pDirection,
                pFlowConnection,
                pFlowConnectionCrossingList );
    }
}
