package de.heinerkuecker.asciiartflowparser;
import de.heinerkuecker.util.StringUtil;

/**
 * Text an einer Verbindung zweier Knoten.
 *
 * Dabei wird der FlowConnectionText wie FlowNode als Rechteck definiert.
 *
 * @author Heiner Kcker
 */
public final class FlowConnectionText
extends FlowConnectionPart
implements Comparable
{
    private final String textStr;

    /**
     * Zeile oben links des belegten Rechteck-Bereiches
     */
    private final int rowLeftTop;

    /**
     * Spalte oben links des belegten Rechteck-Bereiches
     */
    private final int colLeftTop;

    /**
     * Zeile unten rechts des belegten Rechteck-Bereiches
     */
    private final int rowRightBottom;

    /**
     * Spalte unten rechts des belegten Rechteck-Bereiches
     */
    private final int colRightBottom;

    /**
     * Zeile, an welcher die Eintritts-Linie eintritt (eine Position nach dem Ende der Linie)
     */
    private final int inRow;

    /**
     * Spalte, an welcher die Eintritts-Linie eintritt (eine Position nach dem Ende der Linie)
     */
    private final int inCol;

    /**
     * Richtung der eintretenden Linie
     */
    private final int direction;

    /**
     * Ob es bei einem Text vertikal Leerzeilen gibt
     */
    private final boolean hasEmptyRow;

    /**
     * @return the inRow
     */
    /*public*/ int getInRow() {
        return this.inRow;
    }

    /**
     * @return the inCol
     */
    /*public*/ int getInCol() {
        return this.inCol;
    }

    /**
     * @return the direction
     */
    /*public*/ int getDirection()
    {
        return this.direction;
    }
    /**
     * @return the textStr
     */
    public String getText()
    {
        return this.textStr;
    }
    /**
     * @return the textStr
     */
    public String getTextTrimmed()
    {
        return this.textStr.trim();
    }
    /**
     * Constructor.
     *
     * @param pRowLeftTop
     * @param pColLeftTop
     * @param pRowRightBottom
     * @param pColRightBottom
     * @param pNodeText
     */
    private FlowConnectionText(
            final String pTextStr,
            final int pInRow,
            final int pInCol,
            final int pDirection,
            final int pRowLeftTop,
            final int pColLeftTop,
            final int pRowRightBottom,
            final int pColRightBottom,
            final boolean pHasEmptyRow,
            final FlowConnection pFlowConnection )
    {
        super( pFlowConnection );
        this.inRow = pInRow;
        this.inCol = pInCol;
        this.direction = pDirection;
        this.rowLeftTop = pRowLeftTop;
        this.colLeftTop = pColLeftTop;
        this.rowRightBottom = pRowRightBottom;
        this.colRightBottom = pColRightBottom;
        this.textStr = pTextStr;
        this.hasEmptyRow = pHasEmptyRow;
        if ( pTextStr == null )
        {
            throw new NullPointerException();
        }
    }

    /**
     * debug output.
     */
    public String toString()
    {
        return "\n\t\t" + StringUtil.getShortClassName( this ) + " " //--
        + "inRow: " + this.inRow + " " //--
        + "inCol: " + this.inCol + " " //--
        + "direction: " + FlowConnection.getDirectionStr( this.direction ) + " " //--
        + "text: >>" + this.getTextTrimmed() + "<< " //--
        + "rowLeftTop: " + this.rowLeftTop + " " //--
        + "colLeftTop: " + this.colLeftTop + " " //--
        + "rowRightBottom: " + this.rowRightBottom + " " //--
        + "colRightBottom: " + this.colRightBottom;
    }
    /**
     * Zurckgeben der Zeile der
     * nchsten Position, wenn die Linie
     * weiter laufen wrde.
     *
     * @return
     */
    /*public*/ int getNextPosRow()
    {
        // TODO Umstellen auf direction member
        if ( this.direction == FlowConnection.DOWN )
            // nach unten
        {
            // eine Zeile nach unten
            return this.inRow + 1 + ( this.hasEmptyRow ? 2 : 0 );
        }
        else if ( this.direction == FlowConnection.UP )
            // nach oben
        {
            // eine Zeile nach oben
            return this.inRow - 1 - ( this.hasEmptyRow ? 2 : 0 );
        }
        return this.inRow;
    }

    /**
     * Zurckgeben der Spalte der
     * nchsten Position, wenn die Linie
     * weiter laufen wrde.
     *
     * @return
     */
    /*public*/ int getNextPosCol()
    {
        if ( this.direction == FlowConnection.RIGHT )
            // nach rechts
        {
            // eine Spalte nach rechts
            return this.inCol + this.textStr.length();
        }
        else if ( this.direction == FlowConnection.LEFT )
        {
            // eine Spalte nach links
            return this.inCol - this.textStr.length();
        }
        return this.inCol;
    }

    /**
     * @see FlowConnectionPart#matchPosition(int, int)
     */
    /*public*/ boolean matchPosition(
            final int pRowNr,
            final int pColNr)
    {
// TODO umstellen fr mehrzeiligen text
        switch ( this.direction )
        {
            case FlowConnection.UP:
                return //--
                   pRowNr <= this.rowLeftTop
                && pRowNr >= this.rowRightBottom
                && pColNr >= this.colLeftTop
                && pColNr <= this.colRightBottom;
            case FlowConnection.DOWN:
                return //--
                   pRowNr >= this.rowLeftTop
                && pRowNr <= this.rowRightBottom
                && pColNr >= this.colLeftTop
                && pColNr <= this.colRightBottom;
            case FlowConnection.LEFT:
                return //--
                pRowNr == this.inRow //--
                && pColNr <= this.inCol //--
                && pColNr > this.inCol - this.textStr.length();
            case FlowConnection.RIGHT:
                return //--
                    pRowNr == this.inRow //--
                    && pColNr >= this.inCol //--
                    && pColNr < this.inCol + this.textStr.length();
            default:
                throw new IllegalStateException( "unknown direction: " + this.direction );
        }
    }

    /**
     * static methods
     */

    /**
     * parse
     *
     * @param pFlowStrArr
     * @param pInRow
     * @param pInCol
     * @return
     * @throws FlowNetParseException
     */
    /*public*/ static FlowConnectionText parse(
            final String[] pFlowStrArr,
            final int pInRow,
            final int pInCol,
            final int pDirection,
            final FlowConnection pStartFlowConnection )
        throws FlowNetParseException
    {
        /*final*/ String lineStr = pFlowStrArr[ pInRow ];
        if ( pInCol >= lineStr.length() // an der Beginn-Position ist kein Zeichen
                && ( pDirection == FlowConnection.DOWN // eine Zeile unter der Beginn-Position ist kein Zeichen
                        && pInRow + 1 < pFlowStrArr.length //--
                        && pInCol < pFlowStrArr[ pInRow + 1 ].length() ) //--
                && ( pDirection == FlowConnection.UP // eine Zeile ber der Beginn-Position ist kein Zeichen
                        && pInRow > 0 //--
                        && pInCol < pFlowStrArr[ pInRow - 1 ].length() ) )
        {
            // kein FlowConnectionText
            return null;
        }
        int colNr = pInCol;
        int rowNr = pInRow;
        char firstChar = ( pInCol < lineStr.length() ? lineStr.charAt( pInCol ) : 0 );
        if ( ( firstChar == '-' && FlowConnection.isHorizontal( pDirection ) ) //--
            || ( firstChar == '|' && FlowConnection.isVertical( pDirection ) ) //--
            || firstChar == '+' )
        {
            // kein FlowConnectionText
            return null;
        }
        switch ( pDirection )
        {
            case FlowConnection.UP:
            {
                if ( colNr >= pFlowStrArr[ pInRow ].length() // Zeile krzer als die Anfangs-Spalte
                         || ( colNr < pFlowStrArr[ pInRow ].length() // Leerzeichen an der Anfangs-Position
                            && pFlowStrArr[ pInRow ].charAt( colNr ) == ' ' ) )
                        // es folgt ein Leerzeichen
                    {
                        if ( pInRow > 3  //--
                                && ( ( colNr >= pFlowStrArr[ pInRow - 2 ].length() )
                                || ( colNr < pFlowStrArr[ pInRow - 2 ].length() //--
                                && pFlowStrArr[ pInRow - 2 ].charAt( colNr ) == ' ' ) ) )
                            // Prfen auf Leerzeichen nach dem Text oben
                            // es folgt noch ein Leerzeichen (Leerzeile) oben nach dem Text
                        {
//                          if ( ! ( //--
//                                  rowNr > pFlowStrArr.length + 3 //--
//                                  && colNr < pFlowStrArr[ rowNr + 2 ].length() //--
//                                  && pFlowStrArr[ rowNr + 2 ].charAt( colNr ) == ' ' ) )
//                              // Prfen auf Leerzeichen nach dem Text unten
//                          {
//                              return null;
//                          }
                            rowNr--;
                            lineStr = pFlowStrArr[ rowNr ];
                            return parseTextForVerticalDirection( //--
                                    pInRow, //--
                                    pInCol, //--
                                    pDirection, //--
                                    pStartFlowConnection, //--
                                    lineStr, //--
                                    colNr, //--
                                    true // pHasEmptyRow
                                    );
                        }
                    }
                    else
                        // es folgt kein Leerzeichen
                    {
                        return parseTextForVerticalDirection( //--
                                pInRow, //--
                                pInCol, //--
                                pDirection, //--
                                pStartFlowConnection, //--
                                lineStr, //--
                                colNr, //--
                                false // pHasEmptyRow
                                );
                    }
                    return null;
            }
            case FlowConnection.DOWN:
            {
                if ( colNr >= pFlowStrArr[ pInRow ].length() // Zeile krzer als die Anfangs-Spalte
                         || ( colNr < pFlowStrArr[ pInRow ].length() // Leerzeichen an der Anfangs-Position
                            && pFlowStrArr[ pInRow ].charAt( colNr ) == ' ' ) )
                        // es folgt ein Leerzeichen
                    {
                        if ( pInRow + 3 < pFlowStrArr.length //--
                                && ( ( colNr >= pFlowStrArr[ pInRow + 2 ].length() )
                                || ( colNr < pFlowStrArr[ pInRow + 2 ].length() //--
                                && pFlowStrArr[ pInRow + 2 ].charAt( colNr ) == ' ' ) ) )
                            // Prfen auf Leerzeichen nach dem Text unten
                            // es folgt noch ein Leerzeichen (Leerzeile) nach dem Text
                        {
//                          if ( ! ( //--
//                                  rowNr > pFlowStrArr.length + 3 //--
//                                  && colNr < pFlowStrArr[ rowNr + 2 ].length() //--
//                                  && pFlowStrArr[ rowNr + 2 ].charAt( colNr ) == ' ' ) )
//                              // Prfen auf Leerzeichen nach dem Text unten
//                          {
//                              return null;
//                          }
                            rowNr++;
                            lineStr = pFlowStrArr[ rowNr ];
                            return parseTextForVerticalDirection( //--
                                    pInRow, //--
                                    pInCol, //--
                                    pDirection, //--
                                    pStartFlowConnection, //--
                                    lineStr, //--
                                    colNr, //--
                                    true // pHasEmptyRow
                                    );
                        }
                    }
                    else
                        // es folgt kein Leerzeichen
                    {
                        return parseTextForVerticalDirection( //--
                                pInRow, //--
                                pInCol, //--
                                pDirection, //--
                                pStartFlowConnection, //--
                                lineStr, //--
                                colNr, //--
                                false // pHasEmptyRow
                                );
                    }
                    return null;
            }
            case FlowConnection.LEFT:
            {
                while ( colNr  - 2 > 0 //--
                        && ( ! "--".equals( lineStr.substring( colNr - 2 , colNr ) ) ) //--
                        && ( ! "-+".equals( lineStr.substring( colNr - 2 , colNr ) ) ) //--
                        && ( ! "->".equals( lineStr.substring( colNr - 2 , colNr ) ) ) //--
                        )
                {
//System.out.println( lineStr.substring( colNr , colNr + 2 ) );
                    if ( "  ".equals( lineStr.substring( colNr - 2 , colNr ) ) )
                        // zwei oder mehr Leerzeichen
                    {
                        throw new FlowNetParseException( "FlowConnectionText lost at row " + ( pInRow + 1 ) + " col " + ( colNr + 1 ) + " for connection text" );
                    }
                    colNr--;
                }
                if ( colNr > pInCol )
                {
                    // das ist kein FlowConnectionText
                    return null;
                }
//              if ( "|".equals( lineStr.substring( colNr , pCol + 1 ) ) )
//                  // scheinbar eine verirrte Kreuzung
//              {
//                  return null;
//              }
                return new FlowConnectionText( //--
                        lineStr.substring( colNr , pInCol + 1 ), //--
                        pInRow, //--
                        pInCol, //--
                        pDirection, //--
                        pInRow, //--
                        pInCol, //--
                        pInRow, //--
                        colNr, //--
                        false, // hasEmptyRow
                        pStartFlowConnection );
            }
            case FlowConnection.RIGHT:
            {
                while ( colNr < lineStr.length() - 2 //--
                        && ( ! "--".equals( lineStr.substring( colNr , colNr + 2 ) ) ) //--
                        && ( ! "-+".equals( lineStr.substring( colNr , colNr + 2 ) ) ) //--
                        && ( ! "->".equals( lineStr.substring( colNr , colNr + 2 ) ) ) //--
                        )
                {
//System.out.println( lineStr.substring( colNr , colNr + 2 ) );
                    if ( "  ".equals( lineStr.substring( colNr , colNr + 2 ) ) )
                        // zwei oder mehr Leerzeichen
                    {
                        throw new FlowNetParseException( "FlowConnectionText lost at row " + ( pInRow + 1 ) + " col " + ( colNr + 1 ) + " for connection text" );
                    }
                    colNr++;
                }
                if ( colNr <= pInCol )
                {
                    // das ist kein FlowConnectionText
                    return null;
                }
//              if ( "|".equals( lineStr.substring( pCol , colNr ) ) )
//                  // scheinbar eine verirrte Kreuzung
//              {
//                  return null;
//              }
                return new FlowConnectionText( //--
                        lineStr.substring( pInCol , colNr ), //--
                        pInRow, //--
                        pInCol, //--
                        pDirection, //--
                        pInRow, //--
                        pInCol, //--
                        pInRow, //--
                        colNr - 1 , //--
                        false, // hasEmptyRow
                        pStartFlowConnection );
            }
            default:
                throw new IllegalArgumentException( "unknown direction: " + pDirection );
        }
    }

    /**
     * parse Text For Vertical Direction.
     *
     * @param pInRow
     * @param pInCol
     * @param pDirection
     * @param pStartFlowConnection
     * @param pLineStr
     * @param colNr
     * @param pHasEmptyRow
     * @return
     */
    private static FlowConnectionText parseTextForVerticalDirection(
            final int pInRow, //--
            final int pInCol, //--
            final int pDirection, //--
            final FlowConnection pStartFlowConnection, //--
            final String pLineStr, //--
            final int pColNr, //--
            boolean pHasEmptyRow)
    {
        int colNr = pColNr;
        // suchen nach links
        while ( colNr > 0 //--
                && ( colNr > 2 //--
                        && ! "  ".equals( pLineStr.substring( colNr - 2 , colNr ) ) ) //--
                || ( colNr > 1 //--
                        && ! " ".equals( pLineStr.substring( colNr - 1 , colNr ) ) ) //--
                || ( colNr > 0 //--
                        && ! " ".equals( pLineStr.substring( colNr - 1 , colNr ) ) ) //--
                )
        {
            colNr--;
        }
        int startColNr = colNr;
        // suchen nach rechts
        while ( colNr < pLineStr.length() //--
                && ( ! ( colNr + 1 < pLineStr.length() //--
                            && "  ".equals( pLineStr.substring( colNr , colNr + 2 ) ) ) ) //--
                && ( ! ( colNr + 0 < pLineStr.length() //--
                        && " ".equals( pLineStr.substring( colNr , colNr + 1 ) ) ) ) //--
                )
        {
//System.out.println( pColNr + " " + pLineStr.length() );
//System.out.println( "pLineStr.substring( pColNr , pColNr + 2 ): '" + pLineStr.substring( pColNr , pColNr + 2 ) + "'" );
            colNr++;
        }
        if ( colNr <= startColNr )
        {
            // das ist kein FlowConnectionText
            return null;
        }
//                          if ( "-".equals( lineStr.substring( startColNr , colNr ) ) )
//                              // scheinbar eine verirrte Kreuzung
//                          {
//                              return null;
//                          }
        final int emptyRowOffset = ( pHasEmptyRow ? 1 : 0 ) * ( pDirection == FlowConnection.DOWN ? 1 : -1 );

        return new FlowConnectionText( //--
                pLineStr.substring( startColNr , colNr ), // pTextStr
                pInRow, // pInRow
                pInCol, // pInCol
                pDirection, // pDirection
                pInRow + emptyRowOffset, // pRowLeftTop
                startColNr, // pColLeftTop
                pInRow + emptyRowOffset, // pRowRightBottom
                colNr - 1 , // pColRightBottom
                pHasEmptyRow, // hasEmptyRow
                pStartFlowConnection // pFlowConnection
                );
    }

    /**
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    public int compareTo(
            final Object pOther)
    {
        return this.getTextTrimmed().compareTo( ( (FlowConnectionText) pOther ).getTextTrimmed() );
    }
}
