package de.heinerkuecker.asciiartflowparser;

import java.util.*;

import de.heinerkuecker.util.ListHashMap;
import de.heinerkuecker.util.StringUtil;

/**
 * Fluss-Diagramm-Daten-Netz
 * welches aus einem ASCII-Art-Diagramm
 * eingeparst wurde.
 *
 * @author Heiner Kcker
 */
public final class FlowNet
{
    /**
     * Node List
     */
    private final ArrayList flowNodeList = new ArrayList();

    /**
     * Connection List
     */
    private final ArrayList flowConnectionList = new ArrayList();

    /**
     * Clone des Original-Diagrammes
     */
    private String[] flowStrArr;

    /**
     * @param pFlowNode
     */
    /*public*/ void addFlowNode(
            final FlowNode pFlowNode)
    {
        if ( pFlowNode == null )
        {
            throw new NullPointerException();
        }
        // TODO einamaliges Aufadieren (mit Set) absichern
        this.flowNodeList.add( pFlowNode );
    }

    /**
     *
     * @param pFlowConnectionArr
     */
    /*public*/ void addFlowConnection(
            final FlowConnection pFlowConnection)
    {
        if ( pFlowConnection == null )
        {
            throw new NullPointerException();
        }
        // TODO einamaliges Aufadieren (mit Set) absichern
        this.flowConnectionList.add( pFlowConnection );
    }

    /**
     *
     * @param pFlowConnectionArr
     */
    /*public*/ void addFlowConnectionArr(
            final FlowConnection[] pFlowConnectionArr)
    {
        if ( pFlowConnectionArr == null )
        {
            return;
        }
        // TODO einamaliges Aufadieren (mit Set) absichern
        //this.flowConnectionList.addAll( Arrays.asList( pFlowConnectionArr ) );
        for (int i = 0; i < pFlowConnectionArr.length; i++)
        {
            final FlowConnection flowConnection = pFlowConnectionArr[i];
            this.addFlowConnection( flowConnection );
        }
    }

    /**
     * Get FlowNode list.
     * Return a unmodifiableList zum Schutz gegen Seiteneffekte.
     *
     * @return the flowNodeList
     */
    public List getFlowNodeList()
    {
        // unmodifiableList zum Schutz gegen Seiteneffekte
        return Collections.unmodifiableList( this.flowNodeList );
    }

    /**
     * Gib die FlowNode zurck, welche die bergebene
     * Position matcht oder null, wenn kein Match.
     *
     * @param pRow
     * @param pCol
     * @return
     */
    /*public*/ FlowNode getFlowNodeForPosition(
            final int pRow,
            final int pCol)
    {
        for (int i = 0; i < this.flowNodeList.size(); i++)
        {
            final FlowNode flowNode = (FlowNode) this.flowNodeList.get( i );
            if ( flowNode.matchPosition(
                    pRow,
                    pCol ) )
            {
                return flowNode;
            }
        }
        return null;
    }

    /**
     * Parse Flow Net.
     *
     * @param pFlowStrArr
     * @return
     * @throws FlowNetParseException
     */
    public static FlowNet parse(
            final String[] pFlowStrArr)
        throws FlowNetParseException
    {
        final FlowNet flowNet = new FlowNet();

        //Clonen und Rechts Trimmen des String-Array
        flowNet.flowStrArr = StringUtil.rTrimStrArr( pFlowStrArr );

        /**
         * Check for verbotene Whitespaces
         */
        for (int i = 0; i < pFlowStrArr.length; i++)
        {
            final String str = pFlowStrArr[i];
            if ( str == null )
            {
                throw new NullPointerException( "row " + i + " is null" );
            }
            for (int j = 0; j < str.length(); j++)
            {
                final char runChar = str.charAt( j );
                if ( Character.isWhitespace( runChar ) //--
                        && runChar != ' ' )
                {
                    throw new FlowNetParseException("forbidden white space character at row " + ( i + 1 ) + " col " + ( j + 1 ) );
                }
            }
        }

        /**
         * Suche nach einem + - Zeichen als linke
         * obere Ecke eines Knoten (Node).
         */
        for (int rowNr = 0; rowNr < pFlowStrArr.length; rowNr++)
        {
            final String lineStr = pFlowStrArr[rowNr];
            for (int colNr = 0; colNr < lineStr.length(); colNr++)
            {
                if ( lineStr.charAt( colNr ) == '+' )
                {
                    final FlowNode flowNode =
                        FlowNode.parse(
                                pFlowStrArr,
                                rowNr,
                                colNr );
                    if ( flowNode != null )
                        // FlowNode geparst
                    {
//System.out.println( "flowNode: " + flowNode );
                        flowNet.addFlowNode( flowNode );

                    }
                }
            }
        }

        /**
         * Parsen der Verbindungen.
         * Vorher mssen alle Knoten geparst worden sein, damit
         * der End-Knoten der Verbindung bestimmt werden kann.
         */
        final List flowNodeList = flowNet.getFlowNodeList();

        /**
         * Collection fr alle nicht aufgelsten FlowConnectionArrow's
         */
        final ArrayList noEndNodeArrowList = new ArrayList();

        /**
         * Collection fr alle nicht aufgelsten Kreuzungen (FlowConnectionCrossing's) zum Zuordnen der gekreuzten Verbindung
         */
        final ArrayList flowConnectionCrossingList = new ArrayList();

        for (int i = 0; i < flowNodeList.size(); i++)
        {
            final FlowNode flowNode = (FlowNode)  flowNodeList.get( i );

            /**
             * parse die von diesem Knoten abgehenden Verbindungen
             */
            final FlowConnection[] flowConnectionArr =
                FlowConnection.parseFlowConnectionArr(
                        pFlowStrArr,
                        flowNode ,
                        flowNet,
                        noEndNodeArrowList,
                        flowConnectionCrossingList );

            flowNet.addFlowConnectionArr( flowConnectionArr );
        }

        /**
         * Auflsen aller FlowConnectionArrow's, die kein
         * End-Node haben.
         * Dies mssen entweder ge-jointe oder fehlerhafte
         * FlowConnection's sein.
         */
//      final ArrayList flowConnectionList =
//          flowNet.getFlowConnectionList();
        final FlowConnection[] allFlowConnectionArr = flowNet.getFlowConnectionArr();
        noEndNodeArrowListLoop:
        for (int i = 0; i < noEndNodeArrowList.size(); i++)
        {
            final FlowConnectionArrow flowConnectionArrow = (FlowConnectionArrow) noEndNodeArrowList.get( i );
//System.out.println( "flowConnectionArrow: " + flowConnectionArrow );
            final FlowConnection toJoinFlowConnection = flowConnectionArrow.getFlowConnection();
//System.out.println( "toJoinFlowConnection: " + toJoinFlowConnection );

            for (int j = 0; j < allFlowConnectionArr.length; j++)
            {
                final FlowConnection joinInFlowConnection = allFlowConnectionArr[ j ];

                final FlowConnectionPart matchingFlowConnectionPart = //--
                    joinInFlowConnection.getMatchingFlowConnectionPart( //--
                            flowConnectionArrow.getNextPosRow(), //--
                            flowConnectionArrow.getNextPosCol() );

                if ( matchingFlowConnectionPart != null )
                {
//System.out.println( "joinInFlowConnection: " + joinInFlowConnection );
//System.out.println( "matchingFlowConnectionPart: " + matchingFlowConnectionPart );
                    if ( joinInFlowConnection == toJoinFlowConnection )
                    {
                        /**
                         * Abfangen zirkulrer Referenz auf sich selbst.
                         * Diese Massnahme kann aber nicht zirkulrere
                         * Referenzen ber mehrere Part's abfangen.
                         */
                        throw new FlowNetParseException( "circular connection at" //
                                + " row " + ( flowConnectionArrow.getNextPosRow() + 1 ) //--
                                + " col " + ( flowConnectionArrow.getNextPosCol() + 1 ) );
                    }

//System.out.println( "join " + toJoinFlowConnection.getStartFlowNode().getNodeTextTrimmed() + " to " + joinInFlowConnection.getStartFlowNode().getNodeTextTrimmed() );
                    final FlowConnectionJoin flowConnectionJoin = //--
                        new FlowConnectionJoin( //--
                                toJoinFlowConnection, //--
                                joinInFlowConnection, //--
                                matchingFlowConnectionPart, //--
                                flowConnectionArrow.getNextPosRow(), //--
                                flowConnectionArrow.getNextPosCol(), //--
                                flowConnectionArrow.getDirection() );

                    toJoinFlowConnection.addFlowConnectionPart( flowConnectionJoin );
                    continue noEndNodeArrowListLoop;
                }
            }
            throw new FlowNetParseException( "has no end node or join connection to join in: "
                    //+ flowConnectionArrow
                    + "row: " + ( flowConnectionArrow.getNextPosRow() + 1 ) //--
                    + "col: " + ( flowConnectionArrow.getNextPosCol() + 1 ) //--
                    + "direction: " + FlowConnection.getDirectionStr( flowConnectionArrow.getDirection() ) //
                    );
        }

        /**
         * Auflsen aller Kreuzungen (Vermerken der gekreuzten Verbindung)
         *
         * Sinn dieses Codes ist das Prfen, ob eine Kreuzung wirklich eine
         * Kreuzung ist, oder ob es ein Text war.
         *
         * ---|----
         *
         *  |
         *  |
         *  - (Minus)
         *  |
         *  |
         *
         */
        flowConnectionCrossingListLoop:
        for (int i = 0; i < flowConnectionCrossingList.size(); i++)
        {
            final FlowConnectionCrossing flowConnectionCrossing = (FlowConnectionCrossing) flowConnectionCrossingList.get( i );
//System.out.println( "flowConnectionCrossing: " + flowConnectionCrossing );
            final FlowConnection crossingFlowConnection = flowConnectionCrossing.getFlowConnection();
            for (int j = 0; j < allFlowConnectionArr.length; j++)
            {
                // TODO Methode getMatchingFlowConnectionPart benutzen
                final FlowConnection toCrossFlowConnection = allFlowConnectionArr[ j ];

                final FlowConnectionPart matchingFlowConnectionPart = //--
                    toCrossFlowConnection.getMatchingFlowConnectionPart( //--
                            flowConnectionCrossing.getRow(), //--
                            flowConnectionCrossing.getCol() );

                if ( matchingFlowConnectionPart != null //--
                        && matchingFlowConnectionPart instanceof FlowConnectionLinePart )
                {
//System.out.println( "matchingFlowConnectionPart: " + matchingFlowConnectionPart );
//System.out.println( "join " + toJoinFlowConnection.getStartFlowNode().getNodeTexttrim() + " to " + joinInFlowConnection.getStartFlowNode().getNodeTexttrim() );

                    flowConnectionCrossing.setCrossedPart( (FlowConnectionLinePart) matchingFlowConnectionPart );

                    continue flowConnectionCrossingListLoop;
                }
            }
            // TODO jetzt die Kreuzung noch mal als Text einparsen
            final FlowConnectionText flowConnectionText =
                FlowConnectionText.parse(
                        pFlowStrArr,
                        flowConnectionCrossing.getRow(),
                        flowConnectionCrossing.getCol(),
                        flowConnectionCrossing.getDirection(),
                        crossingFlowConnection );

            if ( flowConnectionText != null )
            {
                crossingFlowConnection.replaceCrossingWithText( //--
                        flowConnectionCrossing, //--
                        flowConnectionText );

                continue flowConnectionCrossingListLoop;
            }
            throw new FlowNetParseException( "has no connection line part to cross: " + flowConnectionCrossing );
        }

        /**
         * Missbrauch der Methode FlowConnection#getEndFlowNodeArr()
         * fr zum Prfen auf zirkulre Referenzen.
         */
//      System.out.println( "comment wieder ein" );
        final FlowConnection[] flowConnectionArr = flowNet.getFlowConnectionArr();
        for (int i = 0; i < flowConnectionArr.length; i++)
        {
            final FlowConnection flowConnection = flowConnectionArr[ i ];
            /**
             * Bei einer zirkulren Referenz wirft die aufgerufene
             * Methode eine FlowParseException.
             */
            //flowConnection.getEndFlowNodeArr(new Stack() );
            flowConnection.getFlowConnectionTextArrAndEndNodeValueObjArr( new Stack() );
        }

        return flowNet;
    }

    /**
     * return Array of all FlowConnection's.
     * @return
     */
    private FlowConnection[] getFlowConnectionArr()
    {
        return (FlowConnection[]) this.flowConnectionList.toArray( new FlowConnection[ this.flowConnectionList.size() ] );
    }

    /**
     * Darstellung als Liste von Verbindungen:
     *
     * <pre>
     * A -> B
     * B -> C
     * </pre>
     *
     * @return
     */
//  public String toConnectionListStr()
//  {
////System.out.println( "FlowNet#toConnectionListStr" );
//      final StringBuffer strBuff = new StringBuffer();
//      for (int i = 0; i < this.flowConnectionList.size(); i++)
//      {
//          final FlowConnection flowConnection = (FlowConnection) this.flowConnectionList.get( i );
//          final String[] runConnectionStrArr = flowConnection.toConnectionStrArr();
//          if ( runConnectionStrArr != null )
//          {
//              for (int j = 0; j < runConnectionStrArr.length; j++)
//              {
//                  final String runConnectionStr = runConnectionStrArr[j];
//                  if ( ! StringUtil.empty( runConnectionStr ) )
//                  {
//                      strBuff.append( runConnectionStr.trim() );
//                      if ( i < this.flowConnectionList.size() - 1 //--
//                              || j < runConnectionStrArr.length - 1 )
//                      {
//                          strBuff.append( '\n' );
//                      }
//                  }
//              }
//          }
//      }
//
//      return strBuff.toString();
//  }

    /**
     * Darstellung als Liste von Verbindungen
     * mit Zeilenumbrchen:
     *
     * <pre>
     * A -> B
     * B -> C
     * </pre>
     *
     * @return
     */
    public String getConnectionStrListStr()
    {
//System.out.println( "FlowNet#toConnectionListStr" );
        final StringBuffer strBuff = new StringBuffer();

        final List sortedConnectionList = this.getConnectionStrList();
        Collections.sort( sortedConnectionList );

        int i = 0;
        for (Iterator iter = sortedConnectionList.iterator(); iter.hasNext();)
        {
            final String connectionStr = (String) iter.next();
            strBuff.append( connectionStr );
            if ( i < sortedConnectionList.size() - 1  )
            {
                strBuff.append( '\n' );
            }
            i++;
        }

        return strBuff.toString();
    }

    /**
     * Darstellung als Liste von Verbindungen
     * jeweils eine Verbindung als String ein
     * Set-Entry:
     *
     * <pre>
     * A -> B
     * B -> C
     * </pre>
     *
     * @return
     */
    /*public*/ Set getConnectionSet()
    {
        final HashSet connectionSet = new HashSet();
        connectionSet.addAll( this.getConnectionStrList() );
        return connectionSet;
    }
    /**
     * Darstellung als String-Liste von Verbindungen
     * jeweils eine Verbindung als String ein
     * List-Entry:
     *
     * <pre>
     * A -> B
     * B -> C
     * </pre>
     *
     * @return
     */
    /*public*/ List getConnectionStrList()
    {
        final ArrayList connectionList = new ArrayList();
        for (int i = 0; i < this.flowConnectionList.size(); i++)
        {
            final FlowConnection flowConnection = (FlowConnection) this.flowConnectionList.get( i );
            if ( ! flowConnection.isFromSplit() )
                // connections, die aus einem Split stammen nicht mit auflisten, weil die direkt von einer Node kommenden Verbindungen die gesplitteten Texte und End-Knoten mit auffhren
            {
                final String[] runConnectionStrArr = flowConnection.toConnectionStrArr();
                if ( runConnectionStrArr != null )
                {
                    for (int j = 0; j < runConnectionStrArr.length; j++)
                    {
                        final String runConnectionStr = runConnectionStrArr[j];
                        if ( ! StringUtil.empty( runConnectionStr ) )
                        {
                            connectionList.add( runConnectionStr.trim() );
                        }
                    }
                }
            }
        }

        return connectionList;
    }

    /**
     * Darstellung als FlowConnectionTextArrAndEndNodeValueObj-Liste von Verbindungen:
     *
     * @return FlowConnectionTextArrAndEndNodeValueObj-Liste
     */
    /*public*/ List getConnectionList()
    {
        final ArrayList connectionList = new ArrayList();
        for (int i = 0; i < this.flowConnectionList.size(); i++)
        {
            final FlowConnection flowConnection = (FlowConnection) this.flowConnectionList.get( i );
            if ( ! flowConnection.isFromSplit() )
                // connections, die aus einem Split stammen nicht mit auflisten, weil die direkt von einer Node kommenden Verbindungen die gesplitteten Texte und End-Knoten mit auffhren
            {
                final FlowConnectionTextArrAndEndNodeValueObj[] runConnectionValObjArr = flowConnection.getFlowConnectionTextArrAndEndNodeValueObjArrCatched();
                if ( runConnectionValObjArr != null )
                {
                    for (int j = 0; j < runConnectionValObjArr.length; j++)
                    {
                        final FlowConnectionTextArrAndEndNodeValueObj runConnectionValObj = runConnectionValObjArr[j];
                        connectionList.add( runConnectionValObj );
                    }
                }
            }
        }

        return connectionList;
    }

    /**
     * Zurckgeben einer Fluss(Zustandsbergangs)-Struktur
     * als Array von Value-Objekten mit einer hierarchische Datenstruktur
     * <pre>
     *   Start-FlowNode
     *     |
     *     +-FlowConnection(s)
     *        |
     *        +-FlowConnectionText(e)
     *        +-End-FlowNode
     * </pre>
     * als Value-Objekt-Array.
     *
     * Dies ist notwendig, weil durch das Splitten und Joinen
     * der Connections intern eine kompliziertere Struktur
     * als nur eine Liste der einzelnen Verbindungen
     * vorgehalten wird.
     *
     * @return Value-Objekt-Array mit allen Knoten und den jeweils abgehenden Verbindungen
     *
     * TODO Cachen des return value
     */
    public FlowNodeAndConnectionsValueObj[] getFlowNodeValueObjArr()
    {
        // holen aller Connections mit Start- und End-Node
        final List flowConnectionTextArrAndEndNodeValueObjList = getConnectionList();

        // alle zu der gleichen Start-Node gehrenden Connections auf ein Map-Entry zusammenfassen
        final ListHashMap nodeMap = new ListHashMap();
        for (int i = 0; i < flowConnectionTextArrAndEndNodeValueObjList.size(); i++)
        {
            final FlowConnectionTextArrAndEndNodeValueObj connValObj = (FlowConnectionTextArrAndEndNodeValueObj) flowConnectionTextArrAndEndNodeValueObjList.get( i );
            // Key: Start-Node, Value (List-Entry): Connection
            nodeMap.put( connValObj.startNode , connValObj );
        }

        final ArrayList flowNodeValueObjList = new ArrayList();

        for (Iterator iter = nodeMap.entrySet().iterator(); iter.hasNext();)
        {
            final Map.Entry mapEntry = (Map.Entry) iter.next();

            //ein zusammenfassendes Value-Objekt je Start-Node
            FlowNodeAndConnectionsValueObj flowNodeAndConnectionsValueObj = new FlowNodeAndConnectionsValueObj();

            //Start-Node
            flowNodeAndConnectionsValueObj.startNode = (FlowNode) mapEntry.getKey();

            //an der Start-Node hngende Verbindungen
            final List connValObjList = (List) mapEntry.getValue();

            Collections.sort( connValObjList );

            flowNodeAndConnectionsValueObj.connectionValueObjArr = (FlowConnectionTextArrAndEndNodeValueObj[]) connValObjList.toArray( new FlowConnectionTextArrAndEndNodeValueObj[ connValObjList.size() ] );

            flowNodeValueObjList.add( flowNodeAndConnectionsValueObj );
        }

        Collections.sort( flowNodeValueObjList );

        return (FlowNodeAndConnectionsValueObj[]) flowNodeValueObjList.toArray( new FlowNodeAndConnectionsValueObj[ flowNodeValueObjList.size() ] );
    }

    /**
     * debug output
     */
    public String toString()
    {
        return StringUtil.getShortClassName( this ) + "\n" //--
        + "flowNodeList: " + this.flowNodeList + "\n" //--
        + "flowConnectionList: " + this.flowConnectionList;
    }

    /**
     * Check dieses FlowNet anhand
     * des bergebenen String-Arrays
     * fr Unit-Test.
     *
     * @param pStrings
     * @return
     */
    public boolean checkDiagram(
            final String[] pCheckStrArr
            // boolean pCheckForDoblyEntryInCheckStrArr
            )
    {
        // folgende Logik soll absichern, dass die Struktur auch bei doppelten Eintrgen (siehe flow_12.txt) funktioniert
        final List connectionList = getConnectionStrList();
        for (int i = 0; i < pCheckStrArr.length; i++)
        {
            final String checkStr = pCheckStrArr[i].trim();
            if ( ! StringUtil.empty( checkStr ))
            {
                boolean hasRemoved = connectionList.remove( checkStr );
                if ( ! hasRemoved )
                {
                    //if ( pCheckForDoblyEntryInCheckStrArr )
                    {
                        System.err.println( "missed entry at line " + ( i + 1 ) + " for check: " + checkStr  );
                        return false;
                    }
                }
            }
        }

        if ( ! connectionList.isEmpty() )
        {
            System.err.println( "no matched entrys: " + connectionList );
            return false;
        }
        return true;

//      final HashSet checkSet = new HashSet();
//      for (int i = 0; i < pCheckStrArr.length; i++)
//      {
//          final String checkStr = pCheckStrArr[i].trim();
//          if ( ! StringUtil.empty( checkStr ))
//          {
//              boolean hasAdded = checkSet.add( checkStr );
//              if ( ! hasAdded )
//              {
////                    if ( pCheckForDoblyEntryInCheckStrArr )
////                    {
////                        System.err.println( "doubly entry at line " + ( i + 1 ) + " for check: " + checkStr  );
////                        return false;
////                    }
//              }
//          }
//      }
//
//      final Set connectionSet = toConnectionSet();
//
//      final boolean retVal = checkSet.equals( connectionSet );
//
//      //Voraussetzung fr den folgenden Code-Block ist die unvernderte Weitergabe der Set's
//      if ( ! retVal )
//      {
//          // Ausgabe der Differenz
//          for (int i = 0; i < pCheckStrArr.length; i++)
//          {
//              final String checkStr = pCheckStrArr[i].trim();
//              if ( ! StringUtil.empty( checkStr ))
//              {
//                  boolean wasIn = connectionSet.remove( checkStr );
//                  if ( ! wasIn )
//                  {
//                      connectionSet.add( "missed: " + checkStr );
//                  }
//              }
//          }
//          System.err.println( "diff: " + connectionSet );
//      }
//      return retVal;
    }

    /**
     * Prfen der Bezeichner aller Knoten (FlowNode)
     * auf Eindeutigkeit.
     *
     * (Jeder Zustands-Name darf nur ein einziges
     *  Mal verwendet werden).
     *
     * Bei Verletzung der Eindeutigkeit wird
     * eine FlowNetParseException geworfen, da
     * in diesem Fall das Diagramm nicht verwendet
     * werden kann.
     *
     * @exception FlowNetParseException wenn Eindeutigkeit verletzt
     */
    public void checkNodeIdentifierUniqueness()
        throws FlowNetParseException
    {
        final HashSet nodeNameSet = new HashSet();
        for (int i = 0; i < this.flowNodeList.size(); i++)
        {
            final FlowNode flowNode = (FlowNode) this.flowNodeList.get( i );
            // TODO FlowNode.getNodeTextNormalized()
            //final String flowNodeName = StringUtil.trim( flowNode.getNodeText() );
            final String flowNodeName = flowNode.getNodeTextTrimmed();

            if ( nodeNameSet.contains( flowNodeName ) )
            {
                throw new FlowNetParseException( "non unique flow node identifier: '" + flowNodeName + "' at " + flowNode );
            }
            nodeNameSet.add( flowNodeName );
        }
    }

    /**
     * get FlowConnectionPart auf der
     * bergebenen Position.
     *
     * @param pRowNr
     * @param pColNr
     * @return null, wenn nicht gefunden
     */
    FlowConnectionPart getMatchingFlowConnectionPart(
            final int pRowNr,
            final int pColNr)
    {
        for (int i = 0; i < this.flowConnectionList.size(); i++)
        {
            final FlowConnection flowConnection = (FlowConnection) this.flowConnectionList.get( i );
//System.out.println( "flowConnection: " + flowConnection );
            final FlowConnectionPart matchingFlowConnectionPart = //--
                flowConnection.getMatchingFlowConnectionPart( //--
                        pRowNr, //--
                        pColNr );

            if ( matchingFlowConnectionPart != null )
            {
                return matchingFlowConnectionPart;
            }
        }
        return null;
    }

    /**
     * get FlowConnectionNode auf der
     * bergebenen Position.
     *
     * @param pRowNr
     * @param pColNr
     * @return null, wenn nicht gefunden
     */
    FlowNode getMatchingFlowNode(
            final int pRowNr,
            final int pColNr)
    {
        for (int i = 0; i < this.flowNodeList.size(); i++)
        {
            final FlowNode flowNode = (FlowNode) this.flowNodeList.get( i );

            if ( flowNode.matchPosition( //--
                        pRowNr, //--
                        pColNr ) )
            {
                return flowNode;
            }
        }
        return null;
    }

    /**
     * Prfen auf Zeichen, die nicht eingeparst wurden
     * und demzufolge nicht zum Diagramm gehren.
     *
     * Bei Auftreten nicht geparster Zeichen wird
     * eine FlowNetParseException geworfen, da
     * in diesem Fall das Diagramm nicht verwendet
     * werden kann.
     *
     * @exception FlowNetParseException bei Auftreten nicht geparster Zeichen
     */
    public void checkForUnparsedCharacters()
        throws FlowNetParseException
    {
        for (int rowNr = 0; rowNr < this.flowStrArr.length; rowNr++)
        {
            final String str = this.flowStrArr[rowNr];
            for (int colNr = 0; colNr < str.length(); colNr++)
            {
                final char runChar = str.charAt( colNr );
                if ( ! Character.isWhitespace( runChar ) //--
                        && ( this.getFlowNodeForPosition(rowNr, colNr ) == null //--
                                && this.getMatchingFlowConnectionPart( rowNr, colNr) == null ) )
                {
                    throw new FlowNetParseException("unparsed character '" + runChar + "' at row " + ( rowNr + 1 ) + " col " + ( colNr + 1 ) );
                }
            }
        }
    }

    /**
     * Prfen auf Knoten, die keine Verbindung
     * eingehend oder ausgehend haben
     * und demzufolge nicht zum Diagramm gehren.
     *
     * Bei Auftreten nicht verbundener Knoten wird
     * eine FlowNetParseException geworfen, da
     * in diesem Fall das Diagramm nicht verwendet
     * werden kann.
     *
     * @exception FlowNetParseException bei Auftreten nicht geparster Zeichen
     */
    public void checkForUnconnectedNodes()
        throws FlowNetParseException
    {
        for (int i = 0; i < this.flowNodeList.size(); i++)
        {
            final FlowNode flowNode = (FlowNode) this.flowNodeList.get( i );
            if ( flowNode.outGoingConnectionListIsEmpty() //--
                    && flowNode.inComingConnectionListIsEmpty() )
            {
                throw new FlowNetParseException("unconnected node " + flowNode );
            }
        }
    }

    /**
     * Prfen auf Verbindungen, die nicht
     * genau einen Text besitzen.
     *
     * Bei Auftreten von Verbindungen mit mehr oder
     * weniger als einem Text wird
     * eine FlowNetParseException geworfen, da
     * in diesem Fall das Diagramm nicht verwendet
     * werden kann.
     *
     * @exception FlowNetParseException bei Auftreten von Verbindungen mit nicht genau einem Text
     */
    public void checkForExactlyOneTextPerConnection()
        throws FlowNetParseException
    {
        final FlowNodeAndConnectionsValueObj[] FlowNodeAndConnectionsValObjArr = this.getFlowNodeValueObjArr();

        for (int i = 0; i < FlowNodeAndConnectionsValObjArr.length; i++)
        {
            final FlowNodeAndConnectionsValueObj flowNodeAndConnectionsValueObj = FlowNodeAndConnectionsValObjArr[ i ];
            final FlowConnectionTextArrAndEndNodeValueObj[] flowConnectionTextArrAndEndNodeValueObjArr = flowNodeAndConnectionsValueObj.connectionValueObjArr;

            if ( //flowConnectionTextArrAndEndNodeValueObjArr != null && //--
                    flowConnectionTextArrAndEndNodeValueObjArr.length > 0 )
                // es gibt abgehende Connections
            {
                for (int j = 0; j < flowConnectionTextArrAndEndNodeValueObjArr.length; j++)
                {
                    final FlowConnectionTextArrAndEndNodeValueObj flowConnectionTextArrAndEndNodeValueObj = flowConnectionTextArrAndEndNodeValueObjArr[ j ];

                    if ( flowConnectionTextArrAndEndNodeValueObj.flowConnectionTextArr.length != 1 )
                    {
                        throw new FlowNetParseException( "count of connection text is not exactly one at: " + flowConnectionTextArrAndEndNodeValueObj );
                    }
                }
            }
        }
    }

    /**
     * Prfen auf Verbindungen, die nicht die
     * geforderte Anzahl Texte besitzen.
     *
     * Bei Auftreten von Verbindungen ,die nicht die
     * geforderte Anzahl Texte besitzen, wird
     * eine FlowNetParseException geworfen, da
     * in diesem Fall das Diagramm nicht verwendet
     * werden kann.
     *
     * @param pMinTextCountPerConnection minimale Anzahl Texte
     * @param pMaxTextCountPerConnection maximale Anzahl Texte
     * @exception FlowNetParseException bei Auftreten von Verbindungen mit zuwenigen oder zuvielen Texten
     */
    public void checkMinMaxTextCountPerConnection(
            final int pMinTextCountPerConnection,
            final int pMaxTextCountPerConnection)
        throws FlowNetParseException
    {
        if ( pMinTextCountPerConnection == 1 //--
            && pMaxTextCountPerConnection == 1 )
        {
            checkForExactlyOneTextPerConnection();
            return;
        }
        final FlowNodeAndConnectionsValueObj[] FlowNodeAndConnectionsValObjArr = this.getFlowNodeValueObjArr();

        for (int i = 0; i < FlowNodeAndConnectionsValObjArr.length; i++)
        {
            final FlowNodeAndConnectionsValueObj flowNodeAndConnectionsValueObj = FlowNodeAndConnectionsValObjArr[ i ];
            final FlowConnectionTextArrAndEndNodeValueObj[] flowConnectionTextArrAndEndNodeValueObjArr = flowNodeAndConnectionsValueObj.connectionValueObjArr;

            if ( //flowConnectionTextArrAndEndNodeValueObjArr != null && //--
                    flowConnectionTextArrAndEndNodeValueObjArr.length > 0 )
                // es gibt abgehende Connections
            {
                for (int j = 0; j < flowConnectionTextArrAndEndNodeValueObjArr.length; j++)
                {
                    final FlowConnectionTextArrAndEndNodeValueObj flowConnectionTextArrAndEndNodeValueObj = flowConnectionTextArrAndEndNodeValueObjArr[ j ];

                    final int len = flowConnectionTextArrAndEndNodeValueObj.flowConnectionTextArr.length;
                    if ( len < pMinTextCountPerConnection //--
                            || len > pMaxTextCountPerConnection )
                    {
                        throw new FlowNetParseException( "count of connection text " + len + " is not in range from " + pMinTextCountPerConnection //--
                                + " to " + pMaxTextCountPerConnection //--
                                + "  at: " + flowConnectionTextArrAndEndNodeValueObj );
                    }
                }
            }
        }
    }

    /**
     * Zurckgeben der minimalen Anzahl
     * Texte die alle Verbindungen besitzen.
     */
    public int getMinTextCountPerConnection()
    {
        int minTextCount = -1;
        final FlowNodeAndConnectionsValueObj[] FlowNodeAndConnectionsValObjArr = this.getFlowNodeValueObjArr();

        for (int i = 0; i < FlowNodeAndConnectionsValObjArr.length; i++)
        {
            final FlowNodeAndConnectionsValueObj flowNodeAndConnectionsValueObj = FlowNodeAndConnectionsValObjArr[ i ];
            final FlowConnectionTextArrAndEndNodeValueObj[] flowConnectionTextArrAndEndNodeValueObjArr = flowNodeAndConnectionsValueObj.connectionValueObjArr;

//          if ( //flowConnectionTextArrAndEndNodeValueObjArr != null && //--
//                  flowConnectionTextArrAndEndNodeValueObjArr.length > 0 )
                // es gibt abgehende Connections
            {
                for (int j = 0; j < flowConnectionTextArrAndEndNodeValueObjArr.length; j++)
                {
                    final FlowConnectionTextArrAndEndNodeValueObj flowConnectionTextArrAndEndNodeValueObj = flowConnectionTextArrAndEndNodeValueObjArr[ j ];

                    final int textCount = flowConnectionTextArrAndEndNodeValueObj.flowConnectionTextArr.length;
                    if ( i == 0 && j == 0 )
                        //erster Durchlauf
                    {
                        minTextCount = textCount;
                    }
                    else
                    {
                        minTextCount = Math.min( minTextCount , textCount );
                    }
                }
            }
        }
        return Math.max( 0 , minTextCount );
    }

    /**
     * Zurckgeben der maximalen Anzahl
     * Texte die alle Verbindungen besitzen.
     */
    public int getMaxTextCountPerConnection()
    {
        int maxTextCount = 0;
        final FlowNodeAndConnectionsValueObj[] FlowNodeAndConnectionsValObjArr = this.getFlowNodeValueObjArr();

        for (int i = 0; i < FlowNodeAndConnectionsValObjArr.length; i++)
        {
            final FlowNodeAndConnectionsValueObj flowNodeAndConnectionsValueObj = FlowNodeAndConnectionsValObjArr[ i ];
            final FlowConnectionTextArrAndEndNodeValueObj[] flowConnectionTextArrAndEndNodeValueObjArr = flowNodeAndConnectionsValueObj.connectionValueObjArr;

            if ( //flowConnectionTextArrAndEndNodeValueObjArr != null && //--
                    flowConnectionTextArrAndEndNodeValueObjArr.length > 0 )
                // es gibt abgehende Connections
            {
                for (int j = 0; j < flowConnectionTextArrAndEndNodeValueObjArr.length; j++)
                {
                    final FlowConnectionTextArrAndEndNodeValueObj flowConnectionTextArrAndEndNodeValueObj = flowConnectionTextArrAndEndNodeValueObjArr[ j ];

                    maxTextCount = Math.max( maxTextCount , flowConnectionTextArrAndEndNodeValueObj.flowConnectionTextArr.length );
                }
            }
        }
        return maxTextCount;
    }

    /**
     * Prfen auf Verbindungen, die nicht
     * eindeutig sind.
     *
     * Bei Auftreten von Verbindungen, die nicht
     * eindeutig sind, wird
     * eine FlowNetParseException geworfen, da
     * in diesem Fall das Diagramm nicht verwendet
     * werden kann.
     *
     * @exception FlowNetParseException bei Auftreten von Verbindungen, die nicht eindeutig sind
     */
    public void checkConnectionUniqueness(
            // TODO final String pDefaultText
                 )
        throws FlowNetParseException
    {
        final List connStrList = this.getConnectionStrList();
        final HashSet connStrSet = new HashSet();
        for (int i = 0; i < connStrList.size(); i++)
        {
            final String connStr = (String) connStrList.get( i );
            if ( connStrSet.contains( connStr ) )
            {
                throw new FlowNetParseException( "ambigous connection: " + connStr );
            }
            connStrSet.add( connStr );
        }
    }

    /**
     * Prfen auf Zustands-bergnge, die nicht
     * eindeutig sind.
     *
     * Dabei darf fr eine Anfangsknoten-Verbindungstext-Kombination
     * nur genau ein Endknoten existieren
     * (kein Split nach Text an Verbindung erlaubt).
     *
     * Wichtig ist vor diesem Check der Test auf die Anzahl
     * Texte je Verbindung auf genau 1 oder 0 bis 1, falls
     * default-bergnge (bei fehlendem Text wird der
     * default-Text benutzt) mglich sind.
     *
     * Bei Auftreten von Zustands-bergngen, die nicht
     * eindeutig sind, wird
     * eine FlowNetParseException geworfen, da
     * in diesem Fall das Diagramm nicht verwendet
     * werden kann.
     *
     * @param pDefaultText zu verwendender Default-Text, falls eine Verbindung keinen Text hat.
     *                     wenn dieser Parameter != null ist, wird davon ausgegangen, dass
     *                     Default-Texte erlaubt sind, alternativ nicht erlaubt.
     * @exception FlowNetParseException bei Auftreten von Verbindungen, die nicht eindeutig sind
     */
    public void checkTransitionUniqueness(
            final String pDefaultText )
        throws FlowNetParseException
    {
        final HashSet connTransitionSet = new HashSet();
        final FlowNodeAndConnectionsValueObj[] FlowNodeAndConnectionsValObjArr = this.getFlowNodeValueObjArr();

        for (int i = 0; i < FlowNodeAndConnectionsValObjArr.length; i++)
        {
            final FlowNodeAndConnectionsValueObj flowNodeAndConnectionsValueObj = FlowNodeAndConnectionsValObjArr[ i ];
            final FlowConnectionTextArrAndEndNodeValueObj[] flowConnectionTextArrAndEndNodeValueObjArr = flowNodeAndConnectionsValueObj.connectionValueObjArr;

            if ( //flowConnectionTextArrAndEndNodeValueObjArr != null && //--
                    flowConnectionTextArrAndEndNodeValueObjArr.length > 0 )
                // es gibt abgehende Connections
            {
                for (int j = 0; j < flowConnectionTextArrAndEndNodeValueObjArr.length; j++)
                {
                    final FlowConnectionTextArrAndEndNodeValueObj flowConnectionTextArrAndEndNodeValueObj = flowConnectionTextArrAndEndNodeValueObjArr[ j ];

                    final int textCount = flowConnectionTextArrAndEndNodeValueObj.flowConnectionTextArr.length;
                    if ( textCount > 1 )
                    {
                        throw new FlowNetParseException( "count of connection text is higher tan one " + textCount + " at: " + flowConnectionTextArrAndEndNodeValueObj );
                    }
                    final String connectionText;
                    if ( textCount < 1 )
                    {
                        if ( pDefaultText == null )
                        {
                            throw new FlowNetParseException( "no connection text at: " + flowConnectionTextArrAndEndNodeValueObj );
                        }
                        else
                        {
                            connectionText = pDefaultText;
                        }
                    }
                    else
                    {
                        connectionText = flowConnectionTextArrAndEndNodeValueObj.flowConnectionTextArr[ 0 ].getTextTrimmed();
                    }

                    final TwoStrValObj twoStrValObj = //--
                        new TwoStrValObj(//--
                                flowConnectionTextArrAndEndNodeValueObj.startNode.getNodeTextTrimmed(), //--
                                connectionText );

                    if ( connTransitionSet.contains( twoStrValObj ) )
                    {
                        throw new FlowNetParseException( "ambigous state transition: " + flowConnectionTextArrAndEndNodeValueObj );
                    }
                    connTransitionSet.add( twoStrValObj );
                }
            }
        }
    }

    /**
     * Hilfsklasse fr Methode TODO
     * @author Heiner Kcker
     */
    private static class TwoStrValObj
    {
        final String str1;
        final String str2;
        /**
         * Constructor.
         *
         * @param pStr1
         * @param pStr2
         */
        public TwoStrValObj(
                final String pStr1,
                final String pStr2)
        {
            this.str1 = pStr1;
            this.str2 = pStr2;
        }
        /* (non-Javadoc)
         * @see java.lang.Object#equals(java.lang.Object)
         */
        public boolean equals(Object pOther)
        {
            if ( pOther == null )
            {
                return false;
            }
            if ( pOther == this )
            {
                return true;
            }
            if ( ! ( pOther instanceof TwoStrValObj ) )
            {
                return false;
            }
            final TwoStrValObj other = (TwoStrValObj) pOther;

            // Achtung: nicht null sicher
            return this.str1.equals( other.str1 ) //--
                && this.str2.equals( other.str2 );
        }
        /* (non-Javadoc)
         * @see java.lang.Object#hashCode()
         */
        public int hashCode()
        {
            // Achtung: nicht null sicher
            return ( this.str1.hashCode() + this.str2.hashCode() ) / 2;
        }
        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        public String toString()
        {
            return StringUtil.getShortClassName( this ) + " " //--
                + "str1: " + this.str1
                + "str2: " + this.str2 ;
        }


    }
}
