package de.heinerkuecker.asciiartflowparser;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.heinerkuecker.util.StringUtil;


/**
 * Diese Klasse reprsentiert einen Knoten
 * (Node) eines aus einem ASCII-Art-Diagramm
 * eingeparsten Fluss-Diagrammes (FlowNet).
 * Im Flussdiagramm entspricht ein Knoten
 * einem Zustand.
 *
 * @author Heiner Kcker
 */
public final class FlowNode
{
    private final String nodeText;

    private final int rowLeftTop;

    private final int colLeftTop;

    private final int rowRightBottom;

    private final int colRightBottom;

    private final ArrayList outGoingConnectionList = new ArrayList();

    private final ArrayList inComingConnectionList = new ArrayList();

    /**
     * Constructor.
     *
     * @param pRowLeftTop
     * @param pColLeftTop
     * @param pRowRightBottom
     * @param pColRightBottom
     * @param pNodeText
     */
    private FlowNode(
            final int pRowLeftTop,
            final int pColLeftTop,
            final int pRowRightBottom,
            final int pColRightBottom,
            final String pNodeText)
    {
        super();
        this.rowLeftTop = pRowLeftTop;
        this.colLeftTop = pColLeftTop;
        this.rowRightBottom = pRowRightBottom;
        this.colRightBottom = pColRightBottom;
        this.nodeText = pNodeText;
    }

    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(final Object pOther)
    {
        if ( pOther == null )
        {
            return false;
        }
        if ( pOther == this )
        {
            return true;
        }
        return toString().equals( pOther.toString() );
    }

    /* (non-Javadoc)
     * @see java.lang.Object#hashCode()
     */
    public int hashCode()
    {
        return toString().hashCode();
    }

    /**
     * add an outgoing FlowConnection.
     *
     * @param pFlowConnection FlowConnection to add
     * @return
     * @see java.util.ArrayList#add(java.lang.Object)
     */
    /*public*/ boolean outGoingConnectionListAdd(
            final FlowConnection pFlowConnection)
    {
        return this.outGoingConnectionList.add( pFlowConnection );
    }

    /**
     * has outgoing FlowConnection's ?
     *
     * @return
     * @see java.util.ArrayList#isEmpty()
     */
    public boolean outGoingConnectionListIsEmpty()
    {
        return this.outGoingConnectionList.isEmpty();
    }

    /**
     * Get outgoing FlowConnection List.
     * Return a unmodifiableList zum Schutz gegen Seiteneffekte.
     *
     * @return the outGoingConnectionList
     */
    public List getOutGoingConnectionList()
    {
        // unmodifiableList zum Schutz gegen Seiteneffekte
        return Collections.unmodifiableList( this.outGoingConnectionList );
    }

    /**
     * add an incoming FlowConnection.
     *
     * @param pFlowConnection FlowConnection to add
     * @return
     * @see java.util.ArrayList#add(java.lang.Object)
     */
    /*public*/ boolean inComingConnectionListAdd(
            final FlowConnection pFlowConnection)
    {
        return this.inComingConnectionList.add( pFlowConnection );
    }

    /**
     * has incoming FlowConnection's ?
     *
     * @return
     * @see java.util.ArrayList#isEmpty()
     */
    public boolean inComingConnectionListIsEmpty()
    {
        return this.inComingConnectionList.isEmpty();
    }

    /**
     * debug output.
     */
    public String toString()
    {
        return StringUtil.getShortClassName( this ) + " " //--
        + "nodeText: " + StringUtil.trim( this.nodeText ) + " " //--
        + "rowLeftTop: " + this.rowLeftTop + " " //--
        + "colLeftTop: " + this.colLeftTop + " " //--
        + "rowRightBottom: " + this.rowRightBottom + " " //--
        + "colRightBottom: " + this.colRightBottom;
    }

    /**
     * @return the colLeftTop
     */
    /*public*/ int getColLeftTop() {
        return this.colLeftTop;
    }

    /**
     * @return the colRightBottom
     */
    /*public*/ int getColRightBottom() {
        return this.colRightBottom;
    }

    /**
     * @return the nodeText
     */
    public String getNodeText() {
        return this.nodeText;
    }

    /**
     * @return the nodeText
     */
    public String getNodeTextTrimmed()
    {
        return this.nodeText.trim();
    }

    /**
     * @return the rowLeftTop
     */
    /*public*/ int getRowLeftTop() {
        return this.rowLeftTop;
    }

    /**
     * @return the rowRightBottom
     */
    /*public*/ int getRowRightBottom() {
        return this.rowRightBottom;
    }

    /**
     * ob die bergebene Position
     * von dieser FlowNode berdeckt wird.
     *
     * @param pRow
     * @param pCol
     * @return
     */
    /*public*/ boolean matchPosition(
            final int pRow,
            final int pCol)
    {
        return //--
               pRow >= this.rowLeftTop
            && pRow <= this.rowRightBottom
            && pCol >= this.colLeftTop
            && pCol <= this.colRightBottom;
    }

    /**
     * Parse FlowNode.
     *
     * @param pFlowStrArr ASCII-Art Fluss-Diagramm
     * @param pRowLeftTop Zeile mit '+'-Zeichen
     * @param pColLeftTop Spalte mit '+'-Zeichen
     * @return FlowNode oder null, wenn nicht parseable
     *
     * TODO erlaube '+' - Zeichen am Rand
     */
    /*public*/ static FlowNode parse(
            final String[] pFlowStrArr,
            final int pRowLeftTop,
            final int pColLeftTop)
    {
        /**
         * Suchen nach einer geschlossenen Flche
         */
        /**
         * Suche nach rechts
         */
        int colNr = pColLeftTop + 1;
        String lineStr = pFlowStrArr[ pRowLeftTop ];
        while ( colNr < lineStr.length() //--
                && lineStr.charAt( colNr ) == '-' )
            // laufen ber alle '-'
        {
            colNr++;
        }
        if ( colNr <= pColLeftTop + 1 )
            // es gab nicht mindestens ein '-'
        {
            // kein FlowNode
            return null;
        }
        if ( colNr >= lineStr.length() //--
                || lineStr.charAt( colNr ) != '+' )
            // kein '+' gefunden
        {
            // kein FlowNode
            return null;
        }

        /**
         * End-Spalte vermerken
         */
        final int colRightBottom = colNr;

        /**
         * Suche nach unten
         */
        int rowNr = pRowLeftTop + 1;
        while ( rowNr < pFlowStrArr.length //--
                && colRightBottom < pFlowStrArr[ rowNr ].length() //--
                && pFlowStrArr[ rowNr ].charAt( colRightBottom ) == '|' )
            // laufen ber alle '|'
        {
            rowNr++;
        }

        if ( rowNr <= pRowLeftTop + 1 )
            // es gab nicht mindestens ein '|'
        {
            // kein FlowNode
            return null;
        }
        if ( rowNr >= pFlowStrArr.length //--
                || colNr >= pFlowStrArr[ rowNr ].length() //--
                || pFlowStrArr[ rowNr ].charAt( colNr ) != '+' )
            // kein '+' gefunden
        {
            // kein FlowNode
            return null;
        }

        /**
         * End-Zeile vermerken
         */
        final int rowRightBottom = rowNr;

        /**
         * Kontrolle nach links
         */
        lineStr = pFlowStrArr[ rowRightBottom ];
        for (colNr = pColLeftTop + 1; colNr < colRightBottom; colNr++)
        {
            if ( colNr >= lineStr.length() //--
                    || lineStr.charAt( colNr ) != '-' )
                // Zeilen-Ende erreicht oder kein '-' gefunden
            {
                // kein FlowNode
                return null;
            }
        }

        /**
         * Kontrolle '+' links unten
         */
        if ( pColLeftTop >= pFlowStrArr[ rowRightBottom ].length() //--
                || pFlowStrArr[ rowRightBottom ].charAt( pColLeftTop ) != '+' )
            // kein '+' links unten
        {
            // kein FlowNode
            return null;
        }

        /**
         * Kontrolle nach oben
         */
        for ( rowNr = pRowLeftTop + 1 ; rowNr < rowRightBottom ; rowNr++ )
        {
            if ( pColLeftTop >= pFlowStrArr[ rowNr ].length() //--
                    || pFlowStrArr[ rowNr ].charAt( pColLeftTop ) != '|' )
                // Zeilen-Ende erreicht oder kein '|' gefunden
            {
                // kein FlowNode
                return null;
            }
        }


        /**
         * Lesen des Node-Text
         */
        final StringBuffer flowTextBuff = new StringBuffer();
        for ( rowNr = pRowLeftTop + 1 ; rowNr < rowRightBottom ; rowNr++ )
        {
            lineStr = pFlowStrArr[ rowNr ];

            final String subLine = lineStr.substring( pColLeftTop + 1 , Math.min( colRightBottom, lineStr.length() ) );

            flowTextBuff.append( subLine );

            if ( rowNr < rowRightBottom - 1 )
                // nicht letzte Zeile
            {
                flowTextBuff.append( '\n' );
            }
        }



        final String flowText = flowTextBuff.toString();

        final FlowNode flowNode =
            new FlowNode(
                    pRowLeftTop,
                    pColLeftTop,
                    rowRightBottom,
                    colRightBottom,
                    flowText );
        return flowNode;
    }

}
