package de.heinerkuecker.statetransitionengine;


import java.util.*;


/**
 * Class to override by the the user of the StateTransitionEngine or using
 * without transit condition and without user defined actions before and after
 * changing the state . <br>
 *
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id$
 */
public class DestinationState {

  /**
   * Remark to destination state transit
   */
  private String remark;

  /**
   * number for the position in the order of check with context to perform
   */
  // private int orderPosition;

  /**
   * name of the state to change
   */
  private String destStateName;

  /**
   * user role set with user role names as string
   */
  private Set roleSet;

  /**
   * flag for all permitted
   */
  private boolean bAllPermitted = false;

  /**
   * Constructor
   */
  public DestinationState(
		  final String pDestStateName,
		  final String pRemark )
  {
    if (pRemark == null) {
      throw new NullPointerException();
    }
    this.remark = pRemark;
    if (pDestStateName == null) {
      throw new NullPointerException();
    }
    this.destStateName = pDestStateName;
  }// end constructor

  /**
   * get remark
   */
  public String getRemark() {
    return remark;
  }//end method

/**
* get name of destination state
*/
  public String getDestStateName() {
    return destStateName;
  }//end method

  /**
   * Check condition with overgiven context method to override by the user of the StateTransitionEngine . <br>
   * Notice: perform no side effects to StateContext because this method will called many times
   */
  public boolean isTransitConditionTrue(Object paObjStateContext) {
    return true;
  }// end method

  /**
   * get description of the {@link #isTransitConditionTrue} method
   */
  public String transitConditionDescription() {
    return "no transit condition specified";
  }// end method

  /**
   * This method will performed before the curent state will changed . <br>
   * Is to override by the user of the state transition engine.
   */
  public void beforeStateChanged(Object paObjStateContext) {}

  /**
   * This method will performed after the current state will changed . <br>
   * Is to override by the user of the state transition engine.
   */
  public void afterStateChanged(Object paObjStateContext) {}

  /**
   * add user role string to user role map
   */
  public void addUserRole(String paStrUserRole) {
    if (roleSet == null) {
      roleSet = new TreeSet();
    }
    roleSet.add(paStrUserRole);
  }// end method

  /**
   * get flag for all permitted
   */
  public void setAllPermitted(boolean paBAllPerm) {
    bAllPermitted = paBAllPerm;
  }// end method

  /**
   * check permission for destination state
   */
  public final boolean isUserPermitted(String paStrUserRole) {
    if (bAllPermitted) {
      return true;
    }
    if (paStrUserRole == null) {
      return false;
    }
    return roleSet.contains(paStrUserRole);

  }// end method

  /**
   * debug output
   */
  public String toString() {
    StringBuffer strBuff = new StringBuffer();

    strBuff.append(getClass().getName() + "\n");
    strBuff.append("remark: " + remark + "\n");
    strBuff.append("destStateName: " + destStateName + "\n");
    strBuff.append("transitConditionDescription: " + transitConditionDescription() + "\n");
    strBuff.append("roleSet: " + roleSet + "\n");
    strBuff.append("bAllPermitted: " + bAllPermitted + "\n");
    return strBuff.toString();
  }// end method

  /**
   * debug output
   */
  public String toFormattedString() {
    StringBuffer strBuff = new StringBuffer();

    strBuff.append("       -DestinationState: " + destStateName + "\n");
    strBuff.append("           Remark: " + remark + "\n");
    strBuff.append("           Class : " + getClass().getName() + "\n");
    strBuff.append("           TransitConditionDescription: " + transitConditionDescription() + "\n");
    strBuff.append("           User Roles: " + roleSet + "\n");
    strBuff.append("           All Permitted: " + bAllPermitted + "\n");
    return strBuff.toString();
  }// end method

}// end class
