package de.heinerkuecker.statetransitionengine;


import java.io.*;
import java.util.*;


/**
 *
 * Eine Zustands-bergangs-Maschine, die auf mehrere
 * Session-, User-, Dokument- und so weiter -Zustnde
 * {@link ApplicationState} angewendet werden kann.
 *
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id$
 */
public class StateTransitionEngine {

  /**
   * the state transition map
   */
  private Map stateMap = new TreeMap();

  /**
   * an redundant List with the sorce items in order of appearance
   */
  private List stateList = new ArrayList();

  /**
   * switch for check user role
   */
  private boolean bUserRole = true;

  /**
   * name of the initial state . <br>
   * Wenn dieser Name gesetzt ist, bekommen
   * neu erzeugte ApplicationStates diesen Status.
   */
  private String strInitialStateName;

  /**
   * set initial state for new ApplicationStates
   */
  public void setInitialState(String paStrInitStateName) {
    if (!hasState(paStrInitStateName)) {
      throw new RuntimeException("initial state " + paStrInitStateName + " is unknown");
    }
    strInitialStateName = paStrInitStateName;
  }// end method

  /**
   * get the name of the initial state . <br>
   * returns null if no initial state is defined.
   */
  public String getInitialStateName() {
    return strInitialStateName;
  }// end method

  /**
   * returns wheter user role check is on
   */
  public boolean isUserRoleOn() {
    return bUserRole;
  }// end method

  /**
   * set wheter user role check is on or not
   */
  public void setUserCheck(boolean paBOn) {
    bUserRole = paBOn;
  }// end method

  /**
   * add a new source state to the state transition map
   */
  public void add(SourceState paState) {
    if (stateMap.get(paState.getName()) != null) {
      throw new StateAlreadyDefinedException(paState.getName());
    }
    stateMap.put(paState.getName(), paState);
    stateList.add(paState);
  }// end method

  /**
   * check whether the specified state exists or not
   */
  public boolean hasState(String paStrCheckState) {
    SourceState newCurrentState = (SourceState) stateMap.get(paStrCheckState);

    return !(newCurrentState == null);
  }// end method

  /**
   * perform the event on the state map
   *
   * @param paStrEventName name of the event to performed
   * @param paObjStateContext Object with context to perform the event
   *        (information for conditions to perform the event)
   * @throws InvalidStateException, UnknownEventException
   * @throws NoMatchedDestinationStateException if no matched destination state was found
   * @throws RuntimeException if user role check is on, use method perform(state,role,event,context)
   */
  public void perform(ApplicationState paAppState, String paStrEventName, Object paObjStateContext)
    throws NoMatchedDestinationStateException {
    if (bUserRole) {
      throw new RuntimeException("user role check is on, use method perform(state,role,event,context) with user role string");
    }
    perform(paAppState, paStrEventName, paObjStateContext, null);
  }// end method

  /**
   * perform the event on the state map
   *
   * @param paStrEventName name of the event to performed
   * @param paObjStateContext Object with context to perform the event
   *        (information for conditions to perform the event)
   * @throws InvalidStateException, UnknownEventException
   * @throws NoMatchedDestinationStateException if no matched destination state was found
   */
  public void perform(ApplicationState paAppState, String paStrEventName, Object paObjStateContext, String paStrUserRole)
    throws NoMatchedDestinationStateException {
    // System.out.println("paStrEventName: " + paStrEventName) ;

    SourceState currentState = (SourceState) stateMap.get(paAppState.getCurrentStateName());
    List destinationStateList = currentState.getDestinationStateListForEvent(paStrEventName);

    // System.out.println("destinationStateMap: " + destinationStateMap) ;

    if (destinationStateList == null) {
      throw new UnknownEventException(paStrEventName);
    }

    Iterator destinationStateIterator = destinationStateList.iterator();

    while (destinationStateIterator.hasNext()) {
      DestinationState destState = (DestinationState) destinationStateIterator.next();

      // System.out.println("#perform: destState: " + destState) ;
      // System.out.println("#perform: destState: " + destState) ;
      // System.out.println("#perform: destState.isTransitConditionTrue() " + destState.isTransitConditionTrue(null)) ;
      // System.out.println("#perform: destState.getDestStateName() " + destState.getDestStateName()) ;
      if ((!bUserRole || destState.isUserPermitted(paStrUserRole)) && destState.isTransitConditionTrue(paObjStateContext)) {
        destState.beforeStateChanged(paObjStateContext);
        paAppState.setState(destState.getDestStateName());
        destState.afterStateChanged(paObjStateContext);
        return;// ready
      }
    }
    // no matched DestinationState found
    throw new NoMatchedDestinationStateException("current state: " + paAppState.getCurrentStateName() + " event: "
        + paStrEventName);

  }// end method

  /**
   * Check validity of state map . <br>
   * All destination states must have an equivalent root state
   */
  public boolean check() {

    // Set stateSet = stateMap.entrySet() ;
    // Iterator stateIterator = stateSet.iterator();
    Iterator stateIterator = stateMap.entrySet().iterator();

    while (stateIterator.hasNext()) {
      SourceState state = (SourceState) ((Map.Entry) stateIterator.next()).getValue();
      Iterator eventIterator = state.getEventMap().entrySet().iterator();

      while (eventIterator.hasNext()) {
        StateTransitEvent transEvent = (StateTransitEvent) ((Map.Entry) eventIterator.next()).getValue();

        Iterator destStateIterator = transEvent.getDestinationStateList().iterator();

        if (!destStateIterator.hasNext()) {
          // System.out.println("state transition event " + transEvent + " has no destination states");
          return false;// ready
        }
        while (destStateIterator.hasNext()) {
          DestinationState destState = (DestinationState) destStateIterator.next();

          // System.out.println( "destState: "+destState ) ;
          // System.out.println( "destState.isTransitConditionTrue() "+destState.isTransitConditionTrue(null) ) ;
          // System.out.println( "destState.getDestStateName() "+destState.getDestStateName() ) ;
          if (!hasState(destState.getDestStateName())) {
            // System.out.println("destination state " + destState.getDestStateName() + " has no equivalent root state");
            return false;// ready
          }
        }
      }
    }
    return true;
  }// end method

  /**
   * Check validity of state map and returns an lined (newline separated) string with error messages . <br>
   * All destination states must have an equivalent root state
   */
  public String getCheckMessages() {
    StringBuffer retStrBuff = new StringBuffer();
    // Set stateSet = stateMap.entrySet() ;
    // Iterator stateIterator = stateSet.iterator();
    Iterator stateIterator = stateMap.entrySet().iterator();

    while (stateIterator.hasNext()) {
      SourceState state = (SourceState) ((Map.Entry) stateIterator.next()).getValue();
      Iterator eventIterator = state.getEventMap().entrySet().iterator();

      while (eventIterator.hasNext()) {
        StateTransitEvent transEvent = (StateTransitEvent) ((Map.Entry) eventIterator.next()).getValue();

        Iterator destStateIterator = transEvent.getDestinationStateList().iterator();

        if (!destStateIterator.hasNext()) {
          retStrBuff.append("state transition event " + transEvent + " has no destination states\n");
          // return false;// ready
        }
        while (destStateIterator.hasNext()) {
          DestinationState destState = (DestinationState) destStateIterator.next();

          // System.out.println( "destState: "+destState ) ;
          // System.out.println( "destState.isTransitConditionTrue() "+destState.isTransitConditionTrue(null) ) ;
          // System.out.println( "destState.getDestStateName() "+destState.getDestStateName() ) ;
          if (!hasState(destState.getDestStateName())) {
            retStrBuff.append("destination state " + destState.getDestStateName() + " has no equivalent root state\n");
            // return false;// ready
          }
        }
      }
    }
    return retStrBuff.toString();
  }// end method

  /**
   * Get a Collection with the names of the possible states to change from the current state . <br>
   * This method is planned for showing the destination states in a select box or other choice gui element
   *
   * @throws RuntimeException if user role check is on, use method getPossibleStateToChangeNames(state,role,event,context)
   */

  /* public*/ Collection getPossibleStateToChangeNames(ApplicationState paAppState, Object paObjStateContext) {
    if (bUserRole) {
      throw new RuntimeException("user role check is on, use method getPossibleStateToChangeNames(state,role,context) with user role string");
    }
    return getPossibleStateToChangeNames(paAppState, paObjStateContext, null);
  }// end method

  /**
   * Get a Collection with the names of the possible states to change from the current state . <br>
   * This method is planned for showing the destination states in a select box or other choice gui element
   */
  Collection getPossibleStateToChangeNames(ApplicationState paAppState, Object paObjStateContext, String paStrUserRole) {
    ArrayList retArrList = new ArrayList();

    SourceState currentState = (SourceState) stateMap.get(paAppState.getCurrentStateName());
    Iterator eventIterator = currentState.getEventMap().entrySet().iterator();

    while (eventIterator.hasNext()) {
      StateTransitEvent ev = (StateTransitEvent) ((Map.Entry) eventIterator.next()).getValue();

      Iterator destStateIterator = ev.getDestinationStateList().iterator();

      while (destStateIterator.hasNext()) {
        DestinationState destState = (DestinationState) destStateIterator.next();

        // System.out.println( "destState: "+destState ) ;
        // System.out.println( "destState.isTransitConditionTrue() "+destState.isTransitConditionTrue(null) ) ;
        // System.out.println( "destState.getDestStateName() "+destState.getDestStateName() ) ;
        if ((!bUserRole || destState.isUserPermitted(paStrUserRole)) && destState.isTransitConditionTrue(paObjStateContext)) {
          retArrList.add(destState.getDestStateName());
          // add only the first matched DestinationState to the list
          break;
        }
      }
    }
    // noch was machen, um einen DestinationState nur einmal anzuzeigen
    return retArrList;
  }// end method

  // method fr die events fr die zur Verfgung stehenden Dest states fr eine valuable selectbox [[event],[dest state]]

  /**
   * Get a Collection with the names of the possible states to change from the current state . <br>
   * This method is planned for showing the destination states in a select box or other choice gui element
   *
   * @throws RuntimeException if user role check is on, use method getPossibleStateToChangeNamesAndEventNames(state,role,context)
   */
  public Collection getPossibleStateToChangeNamesAndEventNames(ApplicationState paAppState, Object paObjStateContext) {
    if (bUserRole) {
      throw new RuntimeException("user role check is on, use method getPossibleStateToChangeNamesAndEventNames(state,role,context) with user role string");
    }
    return getPossibleStateToChangeNamesAndEventNames(paAppState,paObjStateContext,null);
  }// end method

  /**
   * Get a Collection with the names of the possible states to change from the current state . <br>
   * This method is planned for showing the destination states in a select box or other choice gui element
   */
  public Collection getPossibleStateToChangeNamesAndEventNames(ApplicationState paAppState, Object paObjStateContext, String paStrUserRole) {
    ArrayList retArrList = new ArrayList();

    SourceState currentState = (SourceState) stateMap.get(paAppState.getCurrentStateName());
    Iterator eventIterator = currentState.getEventMap().entrySet().iterator();

    while (eventIterator.hasNext()) {
      StateTransitEvent ev = (StateTransitEvent) ((Map.Entry) eventIterator.next()).getValue();

      Iterator destStateIterator = ev.getDestinationStateList().iterator();

      while (destStateIterator.hasNext()) {
        DestinationState destState = (DestinationState) destStateIterator.next();

        // System.out.println( "destState: "+destState ) ;
        // System.out.println( "destState.isTransitConditionTrue() "+destState.isTransitConditionTrue(null) ) ;
        // System.out.println( "destState.getDestStateName() "+destState.getDestStateName() ) ;
        if ((!bUserRole || destState.isUserPermitted(paStrUserRole)) && // .
            destState.isTransitConditionTrue(paObjStateContext)) {
          ArrayList stateAndEventArrList = new ArrayList();

          stateAndEventArrList.add(destState.getDestStateName());
          stateAndEventArrList.add(ev.getName());
          retArrList.add(stateAndEventArrList);
          // add only the first matched DestinationState to the list
          break;
        }
      }
    }
    // noch was machen, um einen DestinationState nur einmal anzuzeigen
    return retArrList;
  }// end method

  /**
   * Get a Collection with the names of all states to change from the current state without notice of the context. <br>
   * This method is planned for showing the destination states in a select box or other choice gui element
   */
  public Collection getAllPossibleStateToChangeNames(ApplicationState paAppState) {
    ArrayList retArrList = new ArrayList();

    SourceState currentState = (SourceState) stateMap.get(paAppState.getCurrentStateName());
    Iterator eventIterator = currentState.getEventMap().entrySet().iterator();

    while (eventIterator.hasNext()) {
      StateTransitEvent transEvent = (StateTransitEvent) ((Map.Entry) eventIterator.next()).getValue();

      Iterator destStateIterator = transEvent.getDestinationStateList().iterator();

      while (destStateIterator.hasNext()) {
        DestinationState destState = (DestinationState) destStateIterator.next();

        // System.out.println( "destState: "+destState ) ;
        // System.out.println( "destState.isTransitConditionTrue() "+destState.isTransitConditionTrue(null) ) ;
        // System.out.println( "destState.getDestStateName() "+destState.getDestStateName() ) ;
        retArrList.add(destState.getDestStateName());
      }
    }
    return retArrList;
  }// end method

  /**
   * build state transition map from xml file
   */
//  public void readXmlInputStream(InputStream paXmlInpStream) 
//  {
//    StateTransMapXmlParser.parseStream(this, paXmlInpStream);
//  }

  /**
   * debug output
   */
  public String toString() {
    StringBuffer strBuff = new StringBuffer();

    strBuff.append(getClass().getName() + "\n");

    strBuff.append("stateMap:\n");
    // Iterator stateIterator = stateMap.entrySet().iterator();
    Iterator stateIterator = stateList.iterator();

    while (stateIterator.hasNext()) {
      // SourceState srcState = (SourceState) ((Map.Entry) stateIterator.next()).getValue();
      SourceState srcState = (SourceState) stateIterator.next();

      strBuff.append(srcState.toFormattedString());
    }

    return strBuff.toString();
  }// end method

  /**
   * check whether the state is final (has no StateTransitEvents and DestinationStates) or not.
   */

  /* public*/ boolean isFinal(ApplicationState paAppState) {
    SourceState currentState = (SourceState) stateMap.get(paAppState.getCurrentStateName());
    List destinationStateList = currentState.getDestinationStateList();

    // eventuell dieses Ergebnis cachen
    return (destinationStateList == null || destinationStateList.size() == 0);
  }// end method

  /**
   * check whether the state is final (has no StateTransitEvents and DestinationStates)
   * for the overgiven context or not.
   */

  /* public*/ boolean isFinal(ApplicationState paAppState,Object paObjStateContext) {
    SourceState currentState = (SourceState) stateMap.get(paAppState.getCurrentStateName());
    List destinationStateList = currentState.getDestinationStateList();

    Iterator iter = destinationStateList.iterator();

    while (iter.hasNext()) {
      DestinationState destState = (DestinationState) iter.next();
      boolean bOk = destState.isTransitConditionTrue(paObjStateContext);

      if (bOk) {
        return false;// an matched destination event found
      }
    }// while
    return true; // no matched destination event found
  }// end method

  /**
   * check whether the state is final (has no StateTransitEvents and DestinationStates)
   * for the overgiven context and user role or not.
   */
  /*public*/ boolean isFinal(ApplicationState paAppState,Object paObjStateContext, String paStrUserRole) {
    SourceState currentState = (SourceState) stateMap.get(paAppState.getCurrentStateName());
    List destinationStateList = currentState.getDestinationStateList();
    Iterator iter = destinationStateList.iterator();

    while (iter.hasNext()) {
      DestinationState destState = (DestinationState) iter.next();
      boolean bOk = (destState.isUserPermitted(paStrUserRole) && destState.isTransitConditionTrue(paObjStateContext));

      if (bOk) {
        return false;// an matched destination event found
      }
    }// while
    return true; // no matched destination event found
  }// end method

  /**
   * lists the final state (have no StateTransitEvents and DestinationStates)
   * in an ArrayList.
   */
  public List listFinalStatesNames() {
    ArrayList retList = new ArrayList();
    Iterator stateIterator = stateMap.entrySet().iterator();

    while (stateIterator.hasNext()) {
      SourceState runSrcState = (SourceState) ((Map.Entry) stateIterator.next()).getValue();

      List runDestStateList = runSrcState.getDestinationStateList();

      if (runDestStateList == null || runDestStateList.size() == 0) {
        retList.add(runSrcState.getName());

      }
    }
    // eventuell dieses Ergebnis cachen
    return retList;

    }// end method

  /**
   * check whether the state is final (has no StateTransitEvents and DestinationStates)
   * for the overgiven context or not.
   */
  public List listFinalStatesNames(Object paObjStateContext) {
    ArrayList retList = new ArrayList();
    Iterator stateIterator = stateMap.entrySet().iterator();

    while (stateIterator.hasNext()) {
      SourceState runSrcState = (SourceState) ((Map.Entry) stateIterator.next()).getValue();

      List runDestStateList = runSrcState.getDestinationStateList();

      if (runDestStateList == null || runDestStateList.size() == 0) {

        retList.add(runSrcState.getName());

      }
      else {

        Iterator iter = runDestStateList.iterator();

        boolean bIsFinal = true;

        while (iter.hasNext()) {
          DestinationState destState = (DestinationState) iter.next();

          boolean bOk = destState.isTransitConditionTrue(paObjStateContext);

          if (bOk) {
            bIsFinal = false;// an matched destination event found
            break;
          }
        }// while
        if (bIsFinal) {
          retList.add(runSrcState.getName());// no matched destination event found
        }
      }
    }
    // eventuell dieses Ergebnis cachen
    return retList;

  }// end method

  /**
   * check whether the state is final (has no StateTransitEvents and DestinationStates)
   * for the overgiven context and user role or not.
   */

  public List listFinalStatesNames(Object paObjStateContext, String paStrUserRole) {
    ArrayList retList = new ArrayList();
    Iterator stateIterator = stateMap.entrySet().iterator();

    while (stateIterator.hasNext()) {
      SourceState runSrcState = (SourceState) ((Map.Entry) stateIterator.next()).getValue();

      List runDestStateList = runSrcState.getDestinationStateList();

      if (runDestStateList == null || runDestStateList.size() == 0) {

        retList.add(runSrcState.getName());

      }
      else {

        Iterator iter = runDestStateList.iterator();

        boolean bIsFinal = true;

        while (iter.hasNext()) {
          DestinationState destState = (DestinationState) iter.next();

          boolean bOk = (destState.isUserPermitted(paStrUserRole) && destState.isTransitConditionTrue(paObjStateContext));

          if (bOk) {
            bIsFinal = false;// an matched destination event found
            break;
          }
        }// while
        if (bIsFinal) {
          retList.add(runSrcState.getName());// no matched destination event found
        }
      }
    }
    // eventuell dieses Ergebnis cachen
    return retList;

  }// end method

  /**
   * concatenate second parameter list to first parameter list
   */
  private void concat(List paListFirst , List paListSecond) {
    Iterator iter = paListSecond.iterator();

    while (iter.hasNext()) {
      paListFirst.add(iter.next()) ;
    }// while
  }// end method

}// end class
