package de.heinerkuecker.util;

import java.util.HashMap;
import java.util.Map;

/**
 * A HashMap with int values
 * which are cumulated
 * (add to former value per call of put method).
 *
 * @author Heiner Kcker
 */
public class IntAddHashMap extends HashMap
{

    /**
     * serial version id generated by eclipse
     */
    private static final long serialVersionUID = 2557263538662506458L;

    /**
     * Constructor.
     */
    public IntAddHashMap()
    {
        super();
    }

    /**
     * Constructor.
     * @param pInitialCapacity
     */
    public IntAddHashMap(
            final int pInitialCapacity)
    {
        super(pInitialCapacity);
    }

    /**
     * Constructor.
     * @param pInitMap
     */
    public IntAddHashMap(
            final Map pInitMap)
    {
        super(pInitMap);
    }

    /**
     * Constructor.
     * @param pInitialCapacity
     * @param pLoadFactor
     */
    public IntAddHashMap(
            final int pInitialCapacity,
            final float pLoadFactor)
    {
        super(pInitialCapacity, pLoadFactor);
    }

    /**
     * This method disabled the super.put method
     * for no numeric objects.
     *
     * @see java.util.HashMap#put(java.lang.Object, java.lang.Object)
     */
    public Object put(
            final Object pKey,
            final Object pVal)
    {
        if ( pVal instanceof IntEntry )
        {
            // Typ-Konvertierung, damit putAll mit einer anderen IntAddHashMap funktioniert
            // aus den IntEntry's werden die int values herausgenommen und aufaddiert
            return put( pKey, ( (IntEntry) pVal ).value );
        }
        if ( pVal instanceof Number )
        {
            // add the value of a number object
            return put( pKey, ( (Number) pVal ).intValue() );
        }

        throw new UnsupportedOperationException( "use the method put(Object,int)" );
    }

    /**
     * Add the given int value to
     * the int value on the position
     * according the given key.
     *
     * @see java.util.HashMap#put(java.lang.Object, java.lang.Object)
     */
    public Object put(
            final Object pKey,
            final int pVal)
    {
        // use the super get method or getEntry to avoid conversion to Integer object in this get method
        //final IntEntry oldIntEntry = (IntEntry) super.get( pKey );
        final IntEntry oldIntEntry = this.getEntry( pKey );
        if ( oldIntEntry == null )
        {
            final IntEntry newIntEntry = new IntEntry();
            newIntEntry.value += pVal;
            super.put( pKey, newIntEntry);
            return null;
        }
        oldIntEntry.value += pVal;
        return oldIntEntry;
    }

    /**
     * Increments (add 1)
     * the int value on the position
     * according the given key.
     *
     * @see java.util.HashMap#put(java.lang.Object, java.lang.Object)
     */
    public Object inc(
            final Object pKey )
    {
        return this.put(
                pKey,
                1 );
    }

    /**
     * This method overrites the {@link HashMap#get} method
     * by returning an Integer value with
     * the int value according the given key.
     *
     * @deprecated use the {@link #getEntry(Object)} method to avoid unnecessary object creation
     * @see java.util.HashMap#get(java.lang.Object)
     */
    public Object get(
            final Object pKey)
    {
        final IntEntry intEntry = (IntEntry) super.get(pKey);
        if ( intEntry == null )
        {
            return null;
        }
        return new Integer( intEntry.value );
    }

    /**
     * This method supplemented the {@link HashMap#get} method
     * by returning an IntEntry object with
     * the int value according the given key.
     *
     * @see java.util.HashMap#get(java.lang.Object)
     */
    public IntEntry getEntry(
            final Object pKey)
    {
        final IntEntry intEntry = (IntEntry) super.get(pKey);
        if ( intEntry == null )
        {
            return null;
        }
        return intEntry;
    }

    /**
     * This method supplemented the {@link HashMap#get} method
     * by returning an int value according the given key
     * or the specified default value, if no entry found.
     *
     * @see java.util.HashMap#get(java.lang.Object)
     */
    public int getInt(
            final Object pKey,
            final int pDefaultValue)
    {
        final IntEntry intEntry = (IntEntry) super.get(pKey);
        if ( intEntry == null )
        {
            return pDefaultValue;
        }
        return intEntry.value;
    }

    /**
     * Entry class with an int value.
     */
    public static final class IntEntry
    {
        /**
         * the int value
         */
        public int value = 0;

        /**
         * debug output
         */
        public String toString()
        {
            return "IntEntry: " + this.value;
        }
    }
    /**
     * Test.
     * @param pStrArr
     * TODO Unit Test
     */
    public static void main(
            final String[] pStrArr )
    {
        IntAddHashMap intAddHashMap = new IntAddHashMap();
        intAddHashMap.put( "a" , 1 );
        intAddHashMap.put( "a" , 1 );
        intAddHashMap.put( "a" , 1 );
        System.out.println( intAddHashMap );

        intAddHashMap.putAll( intAddHashMap );
        System.out.println( intAddHashMap );
    }
}
