package de.heinerkuecker.util;

import java.util.HashMap;
import java.util.Map;

/**
 * A HashMap with long values
 * which are cumulated
 * (add to former value per call of put method).
 *
 * @author Heiner Kcker
 */
public class LongAddHashMap extends HashMap
{

    /**
     * serial version id generated by eclipse
     */
    private static final long serialVersionUID = 2557263538662506458L;

    /**
     * Constructor.
     */
    public LongAddHashMap()
    {
        super();
    }

    /**
     * Constructor.
     * @param pInitialCapacity
     */
    public LongAddHashMap(
            final int pInitialCapacity)
    {
        super(pInitialCapacity);
    }

    /**
     * Constructor.
     * @param pInitMap
     */
    public LongAddHashMap(
            final Map pInitMap)
    {
        super(pInitMap);
    }

    /**
     * Constructor.
     * @param pInitialCapacity
     * @param pLoadFactor
     */
    public LongAddHashMap(
            final int pInitialCapacity,
            final float pLoadFactor)
    {
        super(pInitialCapacity, pLoadFactor);
    }

    /**
     * This method disabled the super.put method
     * for no numeric objects.
     *
     * @see java.util.HashMap#put(java.lang.Object, java.lang.Object)
     */
    public Object put(
            final Object pKey,
            final Object pVal)
    {
        if ( pVal instanceof LongEntry )
        {
            // Typ-Konvertierung, damit putAll mit einer anderen LongAddHashMap funktioniert
            // aus den LongEntry's werden die int values herausgenommen und aufaddiert
            return put( pKey, ( (LongEntry) pVal ).value );
        }
        if ( pVal instanceof Number )
        {
            // add the value of a number object
            return put( pKey, ( (Number) pVal ).longValue() );
        }
        throw new UnsupportedOperationException( "use the method put(Object,long)" );
    }

    /**
     * Add the given long value to
     * the long value on the position
     * according the given key.
     *
     * @see java.util.HashMap#put(java.lang.Object, java.lang.Object)
     */
    public Object put(
            final Object pKey,
            final long pVal)
    {
        // use the super get method or getEntry to avoid conversion to Long object in this get method
        //final LongEntry oldLongEntry = (LongEntry) super.get( pKey );
        final LongEntry oldLongEntry = this.getEntry( pKey );
        if ( oldLongEntry == null )
        {
            final LongEntry newLongEntry = new LongEntry();
            newLongEntry.value += pVal;
            super.put( pKey, newLongEntry);
            return null;
        }
        oldLongEntry.value += pVal;
        return oldLongEntry;
    }

    /**
     * Increments (add 1)
     * the long value on the position
     * according the given key.
     *
     * @see java.util.HashMap#put(java.lang.Object, java.lang.Object)
     */
    public Object inc(
            final Object pKey )
    {
        return this.put(
                pKey,
                1 );
    }

    /**
     * This method overrites the {@link HashMap#get} method
     * by returning an Long value with
     * the long value according the given key.
     *
     * @deprecated use the {@link #getEntry(Object)} method to avoid unnecessary object creation
     * @see java.util.HashMap#get(java.lang.Object)
     */
    public Object get(
            final Object pKey)
    {
        final LongEntry longEntry = (LongEntry) super.get(pKey);
        if ( longEntry == null )
        {
            return null;
        }
        return new Long( longEntry.value );
    }

    /**
     * This method supplemented the {@link HashMap#get} method
     * by returning an LongEntry object with
     * the long value according the given key.
     *
     * @see java.util.HashMap#get(java.lang.Object)
     */
    public LongEntry getEntry(
            final Object pKey)
    {
        final LongEntry longEntry = (LongEntry) super.get(pKey);
        if ( longEntry == null )
        {
            return null;
        }
        return longEntry;
    }

    /**
     * This method supplemented the {@link HashMap#get} method
     * by returning an long value according the given key
     * or the specified default value, if no entry found.
     *
     * @see java.util.HashMap#get(java.lang.Object)
     */
    public long getLong(
            final Object pKey,
            final long pDefaultValue)
    {
        final LongEntry longEntry = (LongEntry) super.get(pKey);
        if ( longEntry == null )
        {
            return pDefaultValue;
        }
        return longEntry.value;
    }

    /**
     * Entry class with an long value.
     */
    public static final class LongEntry
    {
        /**
         * the long value
         */
        public long value = 0;

        /**
         * debug output
         */
        public String toString()
        {
            return "LongEntry: " + this.value;
        }
    }

    /**
     * Test.
     * @param pStrArr
     * TODO Unit Test
     */
    public static void main(
            final String[] pStrArr )
    {
        LongAddHashMap longAddHashMap = new LongAddHashMap();
        longAddHashMap.put( "a" , 1 );
        longAddHashMap.put( "a" , 1 );
        longAddHashMap.put( "a" , 1 );
        System.out.println( longAddHashMap );

        longAddHashMap.putAll( longAddHashMap );
        System.out.println( longAddHashMap );
    }
}
