package de.heinerkuecker.xmlbuilder;

//import org.apache.log4j.Logger;


/**
 * Buffer like StringBuffer for byte array.
 *
 * @author Heiner K&uuml;cker
 */
public final class Str8Buff
{
  /**
   * Our Logger
   */
  //private static Logger log = Logger.getLogger(Str8Buff.class);

    private byte[] bArr;

    /** Pointer to the next free position */
    private int size = 0;

    /**
     * Constructor with capacity.
     * Capacitiy must be a value greater zero.
     *
     * @param pCapacity
     */
    public Str8Buff(final int pCapacity)
    {
        this.bArr = new byte[ pCapacity ];
    }

    /**
     * Constructor with specified String.
     * Attention, no side effects safe.
     *
     * @param pStr
     */
    public Str8Buff(final String pStr)
    {
      this( pStr.getBytes() );
    }

    /**
     * Constructor with specified byte array.
     * Attention, no side effects safe.
     *
     * @param pArr
     */
    public Str8Buff(final byte[] pArr)
    {
        this.bArr = pArr;
        this.size = pArr.length;
    }

    /**
     * Constructor with byte array and capacitiy.
     *
     * @param pArr
     * @param pCapacity
     */
    public Str8Buff(//-
        final byte[] pArr, //--
        final int pCapacity)
    {
        this.bArr = new byte[ Math.max( pArr.length , pCapacity ) ];
        System.arraycopy(pArr, 0, this.bArr, 0, pArr.length);
        this.size = pArr.length;
    }

    /**
     * clear this
     */
    public void clear()
    {
      this.size = 0;
    }

    /**
     * @return
     */
    public final int length()
    {
        return this.size;
    }

    /**
     * Set the pointer in this
     * Str8Buff to initial position
     * for reusing this.
     */
    public void reset()
    {
        this.size = 0;
    }

    /**
     * Access to the internal byte array.
     *
     * @return the internal byte array
     */
    public byte[] innerByteArr()
    {
      return this.bArr;
    }


    /**
     * Get the capacity.
     *
     * @return
     */
    public final int capacity()
    {
        return this.bArr.length;
    }

    /**
     * Set the capacity.
     *
     * @param pCapacity new capacitiy
     * @return new capacitiy
     */
    public final int capacity(final int pCapacity)
    {
        if ( pCapacity > this.size)
        {
            byte[] tArr = this.bArr;
            this.bArr = new byte[ pCapacity ];
            System.arraycopy(tArr, 0, this.bArr, 0, tArr.length);

        }

        return this.bArr.length;
    }

    /**
     * Get the byte at the specified position.
     *
     * @param index
     * @return
     */
    public final byte byteAt(final int index)
    {
        //assert index >= size;
        return this.bArr[index];
    }

    /**
     * Get a clone of the saved bytes.
     *
     * @return
     */
    public final byte[] bytes()
    {
        // make a clone
        byte[] retArr = new byte[ this.size ];
        System.arraycopy(this.bArr, 0, retArr, 0, this.size);
        return retArr;
    }

    /**
     * Append a byte at the next free position.
     * Resize the capacity if neccessary.
     *
     * @param pByte byte to append
     */
    public final void append(final byte pByte)
    {
        if (this.size >= this.bArr.length)
        {
            increase();
        }
        this.bArr[ this.size ] = pByte;
        this.size++;
    }

    /**
     * increase capacity.
     */
    private void increase()
    {
      capacity( (this.size + 1) + Math.min(this.size + 1 , 512 * 1024 ) );
    }

    /**
     * Append a byte array at the next free position.
     * Resize the capacity if neccessary.
     *
     * @param pByteArr byte array to append
     */
    public final void append(//--
        final byte[] pByteArr)
    {
        if (this.size + pByteArr.length >= this.bArr.length)
        {
            capacity( (this.size + pByteArr.length + 1) * 2);
        }
        System.arraycopy(pByteArr, 0, this.bArr, this.size, pByteArr.length);
        this.size += pByteArr.length;
    }

    /**
     * Append the specified range from the specified byte array
     * at the next free position.
     * Resize the capacity if neccessary.
     *
     * @param pByteArr byte array to append
     * @param pFromIndex
     * @param pToIndex
     */
    public final void append(
        final byte[] pByteArr ,
        final int pFromIndex ,
        final int pToIndex )
    {
      final int appendSize = pToIndex - pFromIndex;
      if (this.size + appendSize >= this.bArr.length)
      {
        // TODO: der vergrsserungs faktor ist bei grossen datenmenge zu brutal, ab 1 mb begrenzen
          capacity( (this.size + appendSize + 1) * 2);
      }
      System.arraycopy(pByteArr, pFromIndex, this.bArr, this.size, appendSize);
      this.size += appendSize;
    }

    /**
     * Append a byte (as int) at the next free position.
     * Resize the capacity if neccessary.
     *
     * @param pInt byte to append as int
     */
    public void append( final int pInt)
    {
      append( (byte) ( pInt & 0xff ) );
    }

    /**
     *
     */
    public final String toString()
    {
       return new String( bytes() , 0 , this.size );
    }

    /**
     * Replace the specified old sequence with
     * the specified new sequence in the
     * specified byte array.
     *
     * @param pArr
     * @param pOldArr
     * @param pNewArr
     * @return
     * TODO als static methode passt diese methode eigentlich nicht in diese klasse, extra util-klasse machen
     */
    public static final byte[] replace( //--
        final byte[] pArr, //--
        final byte[] pOldArr, //--
        final byte[] pNewArr)
    {
        final int arrLen = pArr.length;
        final Str8Buff buff = new Str8Buff( (int) ( arrLen * 1.5 ) );
        for (int i = 0; i < arrLen; i++)
        {
            if ( Str8.startsWith(pArr, pOldArr, i) )
            {
                buff.append( pNewArr );
                i += pOldArr.length - 1;
            }
            else
            {
                buff.append( pArr[i] );
            }
        }
        return buff.bytes();
    }

    /**
     * Replace the specified old sequence with
     * the specified new sequence in this object.
     *
     * @param pArr
     * @param pOldArr
     * @param pNewArr
     * @return
     */
    public void innerReplace( //--
        final byte[] pOldArr, //--
        final byte[] pNewArr)
    {
      int pos = 0;
      while ( ( pos = indexOf( pOldArr, pos ) ) >= 0 )
      {
        if ( pOldArr.length < pNewArr.length )
        {
          // insert bytes
          insert( pos, pNewArr.length - pOldArr.length );
        }
        else if ( pOldArr.length > pNewArr.length )
        {
          // remove bytes
          delete( pos, pOldArr.length - pNewArr.length );
        }
        // replace on pos for identical length old and new
        System.arraycopy( pNewArr , 0 , this.bArr , pos , pNewArr.length );
        pos += pNewArr.length;
      }
    }

    /**
     * Insert the specified count of bytes
     * in this at the specified position.
     *
     * @param pPos
     * @param pCount
     */
    private void insert(
        final int pPos,
        final int pCount)
    {
      if (pCount < 0)
      {
        throw new IllegalArgumentException( "pCount: " +pCount );
      }
      while (this.size + pCount >= this.bArr.length)
      {
        //TODO einmal richtig statt while, bei while wird mehrmals kopiert
          increase();
      }
      System.arraycopy(this.bArr, pPos, this.bArr, pPos + pCount, this.size - pPos);
      this.size += pCount;
    }

    /**
     * Insert the specified byte array at the
     * specified position.
     *
     * @param pPos
     * @param pByteArr
     */
    public void insert( //--
        final int pPos,
        final byte[] pByteArr)
    {
      insert( pPos, pByteArr.length );
      System.arraycopy(pByteArr, 0 , this.bArr, pPos, pByteArr.length);
    }

    /**
     * Delete the specified count of bytes
     * in this at the specified position.
     *
     * @param pPos
     * @param pCount
     */
    private void delete(
        final int pPos,
        final int pCount)
    {
      if (pCount < 0)
      {
        throw new IllegalArgumentException( "pCount: " +pCount );
      }
      System.arraycopy(this.bArr, pPos + pCount, this.bArr, pPos , this.size - ( pPos + pCount ) );
      this.size -= pCount;
    }

    /**
     * Check whether this starts with the
     * specified byte array content.
     *
     * @param the asked start byte array
     * @return
     */
    public final boolean startsWith(final byte[] pByteArr)
    {
        if ( pByteArr.length > this.size )
        {
            return false;
        }
        return Str8.startsWith(this.bArr, pByteArr);
    }

    /**
     * @param pSize
     */
    public final byte[] leftLenient(final int pSize)
    {
        return Str8.leftLenient(this.bArr, Math.min( pSize , this.size ) );
    }

    /**
     * @param bs
     * @return
     */
    public final int indexOf(final byte[] pByteArr)
    {
        return Str8.indexOf(this.bArr, pByteArr, 0, this.size);
    }

    /**
     * @param
     * @param pFromPos
     * @return
     */
    public final int indexOf( //--
        final byte[] pByteArr, //--
        final int pFromPos)
    {
        return Str8.indexOf(this.bArr, pByteArr, pFromPos, this.size);
    }

    /**
     * Sub operation like String#substring.
     *
     * @param startPos
     * @param endPos
     * @return
     */
    public final byte[] sub(//--
        final int pFromPos, //--
        final int pToPos)
    {
        if ( pToPos > this.size )
        {
            throw new ArrayIndexOutOfBoundsException(pToPos);
        }
        return Str8.sub(this.bArr, pFromPos, pToPos );
    }

    /**
     * Sub operation like String#substring.
     *
     * @param startPos
     * @param endPos
     * @return
     *
     * TODO noch testen
     */
    public final byte[] sub(//--
        final int pFromPos)
    {
      return this.sub( pFromPos, this.size);
    }

    /**
     * Sub operation like String#substring,
     * but with an copy operation in this
     * object.
     *
     * @param pFromPos
     * @param pToPos
     */
    public void innerSub(
        final int pFromPos,
        final int pToPos)
    {
      final int length = pToPos - pFromPos;
      System.arraycopy( this.bArr, pFromPos , this.bArr, 0, length);
      this.size = length;
    }

    /**
     * Sub operation like String#substring,
     * but with an copy operation in this
     * object.
     *
     * @param pStartPos
     */
    public void innerSub(
        final int pFromPos)
    {
      this.innerSub(pFromPos, this.size);

    }

    /**
     * @param oldPos
     * @param LOG_SIZE
     */
    public final byte[] subLenient(//--
        final int pFromPos, //--
        final int pToPos)
    {
        if ( pFromPos >= this.size )
        {
            return new byte[ 0 ];
        }
        return sub( pFromPos , Math.min( this.size , pToPos ) );
    }
    /**
     * @param
     * @return
     */
    public final boolean endsWith(final byte[] pArr)
    {
        return Str8.endsWith(this.bArr, pArr, this.size);
    }

    /**
     * Equals.
     */
    public boolean equals(Object o)
    {
      if (o.getClass() != this.getClass())
      {
        return false;
      }
      final Str8Buff otherBuf = (Str8Buff) o;
      if (this.size != otherBuf.size)
      {
        return false;
      }

      final byte[] thisByteArr = this.bArr;
      final byte[] otherByteArr = otherBuf.bArr;
      for (int i = 0; i < this.size; i++)
      {
        if ( thisByteArr[i] != otherByteArr[i] )
        {
          return false;
        }
      }
      return true;
    }

    /**
     * beim implementiren von equals immer hashCode implementieren.
     * see 'Effektiv Java programmieren', Joshua Bloch , addison-wesley
     *
     * TODO cache the value for more performance
     */
    public int hashCode()
    {
      int result = 17;
      result = 37 * result + this.bArr.length;
      if ( this.bArr.length > 0 )
      {
        result = 37 * result + this.bArr[ 0 ];
      }
      if ( this.bArr.length > 1 )
      {
        result = 37 * result + this.bArr[ 1 ];
      }
      return result;
    }

    /**
     * Test.
     * @param args
     */
    public static void main(String[] args)
    {
      byte[] bArr = "..abcrn..a123rn..rn.".getBytes();
      System.out.println( new String( bArr ) + "\n----------------------------------" );
      bArr = replace(bArr, "rn..".getBytes(), "rn.".getBytes() );
      System.out.println( new String( bArr ) + "\n----------------------------------" );

      bArr = "..abc\r\n..a123\r\n..\r\n.".getBytes();
      System.out.println( new String( bArr ) + "\n----------------------------------" );
      bArr = replace(bArr, "\r\n..".getBytes(), "\r\n.".getBytes() );
      System.out.println( new String( bArr ) + "\n----------------------------------" );

      Str8Buff buff = new Str8Buff( bArr );
      buff.innerReplace( "\r\n..".getBytes(), "\r\n.".getBytes() );
      bArr = buff.bytes();
      System.out.println( new String( bArr ) + "\n----------------------------------" );

      bArr = "0123456789".getBytes();
      buff = new Str8Buff( bArr );
      System.out.println( buff + "\n----------------------------------" );
      buff.innerSub( 1 );
      System.out.println( buff + "\n----------------------------------" );

      bArr = "0123456789".getBytes();
      buff = new Str8Buff( bArr );
      System.out.println( buff + "\n----------------------------------" );
      buff.innerSub( 1 , 3 );
      System.out.println( buff + "\n----------------------------------" );
    }

}
