package de.heinerkuecker.xmlbuilder;

import java.util.*;

/**
 * Klasse zum Ausgeben von XML-Daten als String . <br />
 * Das Format der Ausgabe kann folgendermassen beeinflusst werden:<br />
 * <code>{@link #elementNewline}</code> neue Zeile je Element. <br />
 * <code>{@link #singleTextElementNewline}</code> neue Zeile je Elemente, die nur eine Textzeile enthalten. <br />
 * <code>{@link #elementIndent}</code> Einrckung je Element oder <code>null</null>,
 *              wenn keine Einrckung . <br />
 *
 * Beschleunigung durch Verwendung {@link Str8Buff} statt {@link java.lang.StringBuffer}.
 *
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id:$
 * TODO Encoding fr Attribute < &lt; , & &amp; , ' &apos;
 *          und Text
 * TODO Kommentare
 * TODO CDATA
 * TODO DTD
 */
public class XmlDocument
{

    /**
     * Test
     *
     * @param args
     */
    public static void main(String[] args) {

        // ---------- Aufbau Testdaten ----------
        final Element rootElement = new Element("root", "rootText");
        final XmlDocument xmlDocument = new XmlDocument(rootElement);

        rootElement.addAttribute("att1", "attValue1");
        rootElement.addAttribute("att2", "attValue2");
        rootElement.addAttribute("att3", "attValue3");

        Element subElement1 = new Element("subElement1", "sub1Text");
        rootElement.addElement(subElement1);

        subElement1.addAttribute("att1", "attValue1");
        subElement1.addAttribute("att2", "attValue2");
        subElement1.addAttribute("att3", "attValue3");

        Element subElement2 = new Element("subElement2", "sub2Text");
        rootElement.addElement(subElement2);

        subElement2.addAttribute("att1", "attValue1");
        subElement2.addAttribute("att2", "attValue2");
        subElement2.addAttribute("att3", "attValue3");

        Element subElement2_1 = new Element("subElement2_1", "sub2_1Text");
        subElement2.addElement(subElement2_1);

        Element subElement2_1_1 = new Element("subElement2_1_1", "sub2_1_1Text");
        subElement2_1.addElement(subElement2_1_1);

        Element subElement2_1_2 = new Element("subElement2_1_2", "sub2_1_2Text\nmehrzeilig");
        subElement2_1.addElement(subElement2_1_2);

        Element subElement2_1_3 = new Element("subElement2_1_3", "sub2_1_3Text");
        subElement2_1.addElement(subElement2_1_3);

        Element subElement2_1_4 = new Element("subElement2_1_4", "sub2_1_4Text");
        subElement2_1.addElement(subElement2_1_4);

        // ---------- Ausgabe in unterschiedlichen Styles ----------

        //Jedes Element in einer eigenen Zeile anzeigen
        xmlDocument.elementNewline = false;

        //Auch Elemente, die nur eine Textzeile enthalten, in einer eigenen Zeile anzeigen
        xmlDocument.singleTextElementNewline = true;

        //Elemente nicht einrcken
        xmlDocument.elementIndent = null;

        System.out.println(xmlDocument.build() + "\n\n");

        //--------------

        //Jedes Element in einer eigenen Zeile anzeigen
        xmlDocument.elementNewline = true;

        //Elemente, die nur eine Textzeile enthalten, nicht in einer eigenen Zeile anzeigen
        xmlDocument.singleTextElementNewline = false;

        //Elemente nicht einrcken
        xmlDocument.elementIndent = null;

        System.out.println(xmlDocument.build() + "\n\n");

        //--------------

        //Jedes Element in einer eigenen Zeile anzeigen
        xmlDocument.elementNewline = true;

        //Auch Elemente, die nur eine Textzeile enthalten, in einer eigenen Zeile anzeigen
        xmlDocument.singleTextElementNewline = true;

        //Elemente nicht einrcken
        xmlDocument.elementIndent = null;

        System.out.println(xmlDocument.build() + "\n\n");

        //--------------

        //Jedes Element in einer eigenen Zeile anzeigen
        xmlDocument.elementNewline = true;

        //Elemente, die nur eine Textzeile enthalten, nicht in einer eigenen Zeile anzeigen
        xmlDocument.singleTextElementNewline = false;

        //Elemente mit Tabulator einrcken
        xmlDocument.elementIndent = "\t";

        System.out.println(xmlDocument.build() + "\n\n");

        //--------------

        //Jedes Element in einer eigenen Zeile anzeigen
        xmlDocument.elementNewline = true;

        //Auch Elemente, die nur eine Textzeile enthalten, in einer eigenen Zeile anzeigen
        xmlDocument.singleTextElementNewline = true;

        //Elemente mit Tabulator einrcken
        xmlDocument.elementIndent = "\t";

        System.out.println(xmlDocument.build() + "\n\n");

    }

    /** das Root-Element */
    private Element rootElement;

    /** die XML-Version fr den Header */
    public String xmlVersion = "1.0";

    /** ob Elemente in einer neuen Zeile angezeigt werden sollen */
    public boolean elementNewline;

    /** ob Elemente, die nur eine Textzeile enthalten, in einer neuen Zeile angezeigt werden sollen */
    public boolean singleTextElementNewline;

    /** Einrckung als String oder null, wenn keine Einrckung */
    public String elementIndent;

    /**
     * Schalter fr speichersparend,
     * erlaubt aber nur einmaligen Aufruf
     * der Methode {@link #build()}.
     */
    public boolean memSave;

    /**
     * Konstruktor mit Root-Element
     */
    public XmlDocument(// --
            final Element pRootElement) {
        if (pRootElement == null) {
            throw new NullPointerException("root element expected");
        }
        this.rootElement = pRootElement;
    }

    /**
     * Umwandeln des Dokumenten-Modells in einen XML-String.
     * @return XML-String
     *
     * TODO mal eine Ascii-Art-Ausgabe klarmachen, wie unten und ber den Tabellen-Layouter
     *  |
     *  +-- xxxx
     *  |
     *  +---yyyyy
     *  |   |
     *  |   +---zzz
     */
    public String build() {
        UnsynchStringBuffer strBuff = new UnsynchStringBuffer();
        strBuff.append( "<?xml version=\"" + this.xmlVersion + "\"?>" );
        if (this.elementNewline) {
            strBuff.append("\n");
        }
        buildRecursiv(strBuff, this.rootElement, this.elementIndent != null ? "" : null);
        return strBuff.toString();
    }

    /**
     * rekursive Umwandlung des Dokumenten-Modells in einen XML-String.
     *
     * @param pStrBuff Ausgabe-StringBuffer
     * @param pElement auszugebendes Element
     * @param pElementIndent Einrckung
     */
    private void buildRecursiv(// --
            final UnsynchStringBuffer pStrBuff, // --
            final Element pElement, //
            final String pElementIndent) {

        if (pElementIndent != null) {
            pStrBuff.append(pElementIndent);
        }
        pStrBuff.append('<');
        pStrBuff.append(pElement.name);
        if (pElement.attributeList != null && !pElement.attributeList.isEmpty()) {
            final ArrayList attributeList = pElement.attributeList;
            for (int i = 0; i < attributeList.size(); i++) {
                final Attribute attribute = (Attribute) attributeList.get(i);
                pStrBuff.append(" " + attribute.name + "=\"" + attribute.value
                        + "\"");
            }
        }
        if (pElement.textOrElementList == null
                || pElement.textOrElementList.isEmpty()) {
            // empty element
            pStrBuff.append("/>");
        } else {
            pStrBuff.append('>');
            final ArrayList subElementList = pElement.textOrElementList;
            if (subElementList != null && !pElement.textOrElementList.isEmpty()) {
                String localElementIndent = null;
                if (pElementIndent != null) {
                    localElementIndent = pElementIndent
                            + this.elementIndent;
                }
                if ( ! this.singleTextElementNewline && subElementList.size() == 1
                        && subElementList.get(0) instanceof String) {
                    // es gibt nur ein einziges Text-Element
                    multiLineIndentation(pStrBuff, //--
                            localElementIndent, //--
                            false, //--
                            (String) subElementList.get( 0 ) );

                    pStrBuff.append("</" + pElement.name + ">");
                } else {
                    if (this.elementNewline) {
                        pStrBuff.append("\n");
                    }
                    for (int i = 0; i < subElementList.size(); i++) {
                        if (subElementList.get(i) instanceof String) {
                            final String txtStr = (String) subElementList.get(i);

                            multiLineIndentation(pStrBuff, //--
                                    localElementIndent, //--
                                    true, //--
                                    txtStr);

                            if (this.elementNewline) {
                                pStrBuff.append("\n");
                            }
                        } else {
                            buildRecursiv(pStrBuff, //--
                                    (Element) subElementList.get(i), //--
                                    localElementIndent);
                        }

                    }
                    if (pElementIndent != null) {
                        pStrBuff.append(pElementIndent);
                    }
                    pStrBuff.append("</" + pElement.name + ">");
                }
            }
            if (this.memSave){
                pElement.textOrElementList= null;
            }

        }
        if (this.elementNewline) {
            pStrBuff.append("\n");
        }
    }

    /**
     * Einrcken mehrzeiliger Texte
     * und Schreiben in den bergebenen
     * StringBuffer.
     *
     * @param pStrBuff
     * @param pElementIndent
     * @param pTxtStr
     */
    private static void multiLineIndentation(//--
            final UnsynchStringBuffer pStrBuff, //--
            final String pElementIndent,//--
            final boolean pStartInNewLine, //--
            final String pTxtStr) {
        if (pTxtStr.indexOf( '\n' ) < 0 )
        {
            //einzeiliger Text
            if (pStartInNewLine //--
                    && pElementIndent != null) {
                pStrBuff.append(pElementIndent);
            }
            pStrBuff.append(pTxtStr);
        }
        else
        {
            //mehrzeiliger Text
            //zeilenweise rausschreiben, damit das Einrcken (indent) klappt
            String[] lineArr = lineArray(pTxtStr);
            for (int j = 0; j < lineArr.length; j++) {
                if ( ( pStartInNewLine || j > 0 ) //--
                        && pElementIndent != null) {
                    pStrBuff.append(pElementIndent);
                }
                pStrBuff.append( lineArr[j].trim() );
                if (j < lineArr.length - 1) {
                    //nicht letzte Zeile
                    pStrBuff.append("\n");
                }
            }
        }
    }

    /**
     * Attribut.
     * name=&quot;value&quot;
     */
    public static class Attribute {
        /** Name des Attributes */
        String name;

        /** Inhalt des Attributes */
        String value;

        /**
         * Konstruktor public.
         *
         * @param pName
         * @param pValue
         */
        public Attribute(// --
                final String pName, // --
                final String pValue) {
            this.name = pName;
            this.value = pValue;
        }
    }

    /**
     * XML-Element.
     * Ein Element kann beliebig viele {@link Attribute},
     * beliebig viele Texte und beliebig viele
     * Unter-Elemente enthalten.
     *
     */
    public static class Element {

        /** Name des Elementes */
        String name;

        /** Liste mit {@link Attribute}n */
        ArrayList attributeList = new ArrayList();

        /** Liste mit Texten (String) oder Unter-Elementen */
        ArrayList textOrElementList = new ArrayList();

        /**
         * Konstruktor public.
         */
        public Element(// --
                final String pName) {
            if (pName == null) throw new NullPointerException("name");
            this.name = pName;
        }

        /**
         * Konstruktor public.
         */
        public Element(// --
                final String pName, // --
                final String pText) {
            if (pName == null) throw new NullPointerException("name");
            if (pText == null) throw new NullPointerException("text");
            this.name = pName;
            this.addText(pText);
        }

        /**
         * add {@link Attribute}.
         *
         * @param pAttribute zu addierendes {@link Attribute}
         */
        public void addAttribute(// --
                final Attribute pAttribute) {
            if (pAttribute == null) throw new NullPointerException("attribute");
            this.attributeList.add(pAttribute);
        }

        /**
         * add {@link Attribute} als Name-Value-Paar.
         *
         * @param pName Name
         * @param pValue Value
         */
        public void addAttribute(// --
                final String pName, // --
                final String pValue) {
            if (pName == null) throw new NullPointerException("name");
            if (pValue == null) throw new NullPointerException("value");
            final Attribute attribute = new Attribute(pName, pValue);
            this.attributeList.add(attribute);
        }

        /**
         * add Element.
         *
         * @param pElement
         */
        public void addElement(// --
                final Element pElement) {
            if (pElement == null) throw new NullPointerException("element");
            this.textOrElementList.add(pElement);
        }

        /**
         * add ein Elemnt mit nur einem Text.
         *
         * @param pName
         * @param pText
         */
        public void addElement(// --
                final String pName, //--
                final String pText) {
            if (pName == null) throw new NullPointerException("name");
            if (pText == null) throw new NullPointerException("text");
            final Element subElem = new Element( pName );
            subElem.addText( pText );
            this.textOrElementList.add(subElem);
        }

        /**
         * add Text . <br/>
         * Texte werden als String-Objekte aufaddiert.
         *
         * @param pText zu addierender Text.
         */
        public void addText(final String pText) {
            if (pText == null) throw new NullPointerException("text");
            this.textOrElementList.add(pText);
        }
    }

    /**
     * zeilenweises Zerlegen des bergebenen
     * Strings.
     *
     * @param pStr zeilenweise zu zerlegender String
     * @return String-Array mit den einzelnen Zeilen
     */
    public static String[] lineArray(final String pStr){
        final StringTokenizer strTok = new StringTokenizer( pStr , "\n" );
        final ArrayList list = new ArrayList();
        while (strTok.hasMoreTokens()){
            list.add( strTok.nextToken());
        }
        return (String[]) list.toArray( new String[ list.size() ] );
    }

}
